'''GPIBWrap.py - A Python Interface to the GPIBlib Instrument Library.
   ===================================================================

   The combination of Python, the instrument-object based C++ GPIBlib
   instrument control library, NumPy and MatPlotLib provides a quite
   useful way of building measurement applications for computer controlled
   instruments.

   Currently implemented instruments are:

   HP_3325A Synthesiser/Function Generator
   HP_3457A Multimeter
   HP_3582A FFT Spectrum Analyser
   HP_8903E Distortion Analyser
   TEK_710A Transient Digitizer (partially)

   Author: Nick Glazzard 2012'''

from ctypes import *
from numpy import *
import matplotlib.pyplot as plt

lib = cdll.LoadLibrary( './GPIBlib.so' )

def argerr():
    raise RuntimeError( 'Invalid arguments.' )

def callerr():
    raise RuntimeError( 'GPIBlib call failed.' )

def eu(v,u):
    exp = { 'n':1e-9, 'u':1e-6, 'm':1e-3, 'k':1e3, 'M':1e6, 'G':1e9 }
    print u[0:1]
    try:
        mult = exp[u[0:1]]
    except Exception, err:
        mult = 1
    return( v * mult )

def frequnits(fu):
    funits = { 'Hz':1, 'kHz':2, 'MHz':3, 'GHz':4 }
    return funits[fu]

def ampunits(au):
    aunits = { 'v':1, 'mv':2, 'v_pp':3, 'mv_pp':4, 'v_rms':5, 'mv_rms':6, 'dbm':7 }
    return aunits[au]

def chancodes(ch):
    ccodes = { 'a':1, 'cha':1, 'b':2, 'chb':2, 'ch1':1, 'ch2':2, 'both':3, 'chboth':3 }
    return ccodes[ch]

def couplecodes(cu):
    cupcodes = { 'ac':1, 'dc':2, 'gnd':3, 'ac_lf_reject':4, 'ac_hf_reject':5 }
    return cupcodes[cu]

class GPIBInst(object):
    '''Base class for GPIB instruments with functions common to all instruments.'''
    def __init__(self):
        self.obj = None

    def set_obj(self,obj):
        self.obj = obj

    def set_print_error(self,b):
        '''Allow (b=1) or suppress (b=0) error messages.'''
        return lib.GpibInstrument_set_print_error(self.obj,c_int(b))
        
    def set_print_gpib(self,b):
        '''Log (b=1) or not (b=0) all raw GPIB traffic.'''
        return lib.GpibInstrument_set_print_gpib(self.obj,c_int(b))
        
    def set_display_status(self,b):
        '''Show status (b=1) or not (b=0) after each GPIB command.'''
        return lib.GpibInstrument_set_display_status(self.obj,c_int(b))

    def set_dump_gpib(self,b):
        '''Dump a representation (b=1) or not (b=0) of binary GPIB read data.'''
        return lib.GpibInstrument_set_dump_gpib(self.obj,c_int(b))

    def reset_bus(self):
        '''Force a reset on the GPIB bus.'''
        return lib.GpibInstrument_reset_bus(self.obj)


class HP_3457A(GPIBInst):
    '''HP 3457A Multimeter Object.'''
    def __init__(self,a,g):
        super(HP_3457A,self).__init__()
        self.obj = lib.HP_3457A_new( a, g )
        super(HP_3457A,self).set_obj(self.obj)
        
    def check_open(self):
        '''Check the device is open. Return True if it is.'''
        return lib.HP_3457A_check_open(self.obj)

    def set_fast_mode(self,b):
        '''Start using the HP 3457A for 3.5 digit measurements as fast as possible.
        If b=1, apply fast mode to AC measurements also (Only recommended for signals >400Hz).'''
        if( lib.HP_3457A_set_fast_mode(self.obj,c_int(b)) != 1 ):
            callerr()
        return True

    def clear_fast_mode(self):
        '''End fast mode. Revert to high precision mode.'''
        if( lib.HP_3457A_clear_fast_mode(self.obj) != 1 ):
            callerr()
        return True

    def set_function(self,f,v):
        '''Set Function to f and Range to v.
        f may be: 'dcv', 'acv', 'acdcv', 'ohm', 'ohm4', 'dci', 'aci', 'acdci', 'freq', 'period'
        v depends on f.
        For volts functions: 'auto', '30mv', '300mv', '3v', '30v', '300v'
        For ohm functions: 'auto', '30r', '300r', '3k', '30k', '300k', '3m', '30m', '3g'
        For time functions: 'auto' '''
        funcs = { 'dcv':1, 'acv':2, 'acdcv':3, 'ohm':4, 'ohm4':5,
                  'dci':6, 'aci':7, 'acdci':8, 'freq':9, 'period':10 }
        vvals = { 'auto':-1, '30mv':0.03, '300mv':0.3, '3v':3, '30v':30, '300v':300 }
        ovals = { 'auto':-1, '30r':30, '300r':300, '3k':3000, '30k':30000, '300k':300000,
                  '3m':3000000, '30m':30000000, '3g':3000000000 }
        ivals = { 'auto':-1, '300u':0.0003, '3m':0.003, '30m':0.03, 
                  '300m':0.3, '1a':1, '1a5':1.5 }
        autos = { 'auto':-1 }
        funcvals = { 1:vvals, 2:vvals, 3:vvals, 4:ovals, 5:ovals, 6:ivals,
                     7:ivals, 8:ivals, 9:autos, 10:autos }
        try:
            code = funcs[f]
            valuetab = funcvals[code]
            value = valuetab[v]
        except Exception, err:
            argerr()
        if( lib.HP_3457A_set_function(self.obj,c_int(code),c_double(value)) != 1 ):
            callerr()
        return True

    def read(self):
        '''Read the last measurement from the HP 3457A.'''
        lib.HP_3457A_read_simple.restype = c_double
        return lib.HP_3457A_read_simple(self.obj)


class HP_3325A(GPIBInst):
    '''HP 3325A Synthesiser/Function Generator Object.'''
    def __init__(self,a,g):
        super(HP_3325A,self).__init__()
        self.obj = lib.HP_3325A_new( a, g )
        super(HP_3325A,self).set_obj(self.obj)

    def check_open(self):
        '''Check the device is open. Return True if it is.'''
        return lib.HP_3325A_check_open(self.obj)

    def check_error(self):
        '''Check the last HP_3325A specific command executed OK. Return True if so.'''
        return lib.HP_3325A_check_error(self.obj)

    def set_function(self,sw):
        '''Set the output waveform to sw.
        sw may be: 'dc', 'sine', 'square', 'triangle', 'ramp_up', 'ramp_down' '''
        funcs = { 'dc':0, 'sine':1, 'square':2, 'triangle':3, 'ramp_up': 4, 'ramp_down':5 }
        try:
            w = funcs[sw]
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_function(self.obj,c_int(w))

    def set_frequency(self,freq,sunits):
        '''Set the output frequency to freq (in sunits).
        sunits may be: 'Hz', 'kHz', 'MHz' '''
        try:
            units = frequnits(sunits)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_frequency(self.obj,c_float(freq),c_int(units))

    def set_amplitude(self,amp,sunits):
        '''Set the output amplitude to amp (in sunits).
        sunits may be: 'v', 'mv', 'v_pp', 'mv_pp', 'v_rms', 'mv_rms', 'dbm' '''
        try:
            units = ampunits(sunits)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_amplitude(self.obj,c_float(amp),c_int(units))

    def set_dc_offset(self,dc,sunits):
        '''Set the output DC offset to dc (in sunits).
        sunits may be: 'v', 'mv' '''
        try:
            units = ampunits(sunits)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_dc_offset(self.obj,c_float(dc),c_int(units))

    def set_phase(self,degrees):
        '''Set the phase offset to degrees degrees.'''
        return lib.HP_3325A_set_phase(self.obj,c_float(degrees))

    def set_sweep_start_frequency(self,freq,sunits):
        '''Set the frequency at the start of a sweep (in sunits).
        sunits may be: 'Hz', 'kHz', 'MHz' '''
        try:
            units = frequnits(sunits)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_sweep_start_frequency(self.obj,c_float(freq),c_int(units))

    def set_sweep_end_frequency(self,freq,sunits):
        '''Set the frequency at the end of a sweep (in sunits).
        sunits may be: 'Hz', 'kHz', 'MHz' '''
        try:
            units = frequnits(sunits)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_sweep_end_frequency(self.obj,c_float(freq),c_int(units))

    def set_sweep_marker_frequency(self,freq,sunits):
        '''Set the frequency at which to output the marker signal during a sweep (in sunits).
        sunits may be: 'Hz', 'kHz', 'MHz' '''
        try:
            units = frequnits(sunits)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_sweep_marker_frequency(self.obj,c_float(freq),c_int(units))

    def set_sweep_mode(self,ssweep_mode):
        '''Set the sweep mode to 'lin' or 'log'. '''
        smodes = { 'lin':1, 'log':2 }
        try:
            sweep_mode = smodes(ssweep_mode)
        except Exception, err:
            argerr()
        return lib.HP_3325A_set_sweep_mode(self.obj,c_int(sweep_mode))

    def set_sweep_time(self,secs):
        '''Set the time over which to do a sweep (in seconds).'''
        return lib.HP_3325A_set_sweep_time(self.obj,c_float(secs))

    def start_sweep_single(self,waitfin):
        '''Do a single sweep. If waitfin=1, wait for it to end.'''
        return lib.HP_3325A_start_sweep_single(self.obj,c_int(waitfin))

    def start_sweep_continuous(self):
        '''Start sweeping continuously.'''
        return lib.HP_3325A_start_sweep_continuous(self.obj)

    def use_amplitude_modulation(self,yes):
        '''Turn amplitude modulation on (1) or off (0).'''
        return lib.HP_3325A_use_amplitude_modulation(self.obj,c_int(yes))

    def use_phase_modulation(self,yes):
        '''Turn phase modulation on (1) or off (0).'''
        return lib.HP_3325A_use_phase_modulation(self.obj,c_int(yes))
        

class HP_8903E(GPIBInst):
    '''HP 8903E Distortion Analyzer Object.'''
    def __init__(self,a,g):
        super(HP_8903E,self).__init__()
        self.obj = lib.HP_8903E_new( a, g )
        super(HP_8903E,self).set_obj(self.obj)

    def check_open(self):
        '''Check the device is open. Return True if it is.'''
        return lib.HP_8903E_check_open(self.obj)

    def get_ac_volts(self):
        '''Measure the AC voltage present at the input.'''
        lib.HP_8903E_get_ac_volts.restype = c_double
        return lib.HP_8903E_get_ac_volts(self.obj)

    def get_distortion(self):
        '''Measure the distortion of the signal present at the input.'''
        lib.HP_8903E_get_distortion.restype = c_double
        return lib.HP_8903E_get_distortion(self.obj)


class HP_3582A(GPIBInst):
    '''HP 3582A Spectrum Analyzer Object.'''
    def __init__(self,a,g):
        super(HP_3582A,self).__init__()
        self.obj = lib.HP_3582A_new( a, g )
        super(HP_3582A,self).set_obj(self.obj)
        
    def check_open(self):
        '''Check the device is open. Return True if it is.'''
        return lib.HP_3582A_check_open(self.obj)

    def set_input_mode(self,c):
        '''Set which channels to use. 
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2', 'both', 'chboth' '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        return lib.HP_3582A_set_input_mode(self.obj,c_int(ic))

    def coupling(self,c,m):
        '''Set the coupling mode for channel c to mode m.
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2', 'both', 'chboth'
        m may be: 'ac':1, 'dc' '''
        try:
            ic = chancodes(c)
            im = couplecodes(m)
        except Exception, err:
            argerr()
        return lib.HP_3582A_coupling(self.obj,c_int(ic),c_int(im))

    def set_sensitivity(self,c,v):
        '''Set the sensitivity of channel c to v volts rms.
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2', 'both', 'chboth'
        v is a voltage in the range 0.0 (30mV lowest range) to 30.0. '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        return lib.HP_3582A_set_sensitivity(self.obj,c_int(ic),c_double(v))

    def set_sensitivity_db(self,c,v):
        '''Set the sensitivity of channel c to v volts rms in dB wrt 1V.
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2', 'both', 'chboth'
        v is dBV in the range -50.0 to 30.0. '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        return lib.HP_3582A_set_sensitivity_db(self.obj,c_int(ic),c_double(v))

    def set_frequency(self,m,span,freq):
        '''Set the frequency axis of the spectrum. The use of span and freq depends on mode m.
        'full': 0->25kHz [span] and [freq] ignored.
        'zero_start: 0->[span], [freq] ignored.
        'set_start': [freq]->[freq]+[span]
        'set_center: [freq]-[span]/2 -> [freq]+[span]/2 '''
        fmodes = { 'full':1, 'zero_start':2, 'set_start':4, 'set_center':3 }
        try:
            im = fmodes(m)
        except Exception, err:
            argerr()
        return lib.HP_3582A_set_frequency(self.obj,c_int(im),c_double(span),c_double(freq))

    def display(self,c,ym):
        '''Select which channel(s) to display (c) and the Y scaling mode (ym).
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2', 'both', 'chboth'
        ym may be: 'linear', '10dB', '2dB' (per division). '''
        ymodes = { 'linear':1, '10dB':2, '10db':2, '2dB':3, '2db':3 }
        try:
            ic = chancodes(c)
            iym = ymodes[ym]
        except Exception, err:
            argerr()
        return lib.HP_3582A_display(self.obj,c_int(ic),c_int(iym))

    def set_amp_ref_level_db(self,r):
        '''Set the display full scale point to be r dB below input full scale.
        r will be forced to be a multiple of 10dB. '''
        ir = float( int(r/10.0) * 10 )
        return lib.HP_3582A_set_amp_ref_level_db(self.obj,c_double(ir))

    def get_amp_ref_level_db(self):
        '''Get the display full scale point in dB.'''
        lib.HP_3582A_get_amp_ref_level_db.restype = c_double
        return lib.HP_3582A_get_amp_ref_level_db(self.obj)

    def window(self,w):
        '''Set the window function to w.
        w may be: 'flattop', 'hanning', 'uniform'. '''
        wmodes = { 'flattop':1, 'hanning':2, 'uniform':3 }
        try:
            iw = wmodes[w]
        except Exception, err:
            argerr()
        return lib.HP_3582A_window(self.obj,c_int(iw))

    def start_measurement(self,a,c,m):
        '''Start a measurement. The averaging mode is a: 'off', 'rms'.
        The number of spectra to average is c. This will be made a power of 2 internally.
        The maximum time to wait for completion is m seconds (rounded up to a multiple of 5).
        N.B. Call this BEFORE get_spectrum().'''
        amodes = { 'off':1, 'rms':2 }
        try:
            ia = amodes[a]
        except Exception, err:
            argerr()
        im = int(m)/5 + 1
        return lib.HP_3582A_start_measurement(self.obj,c_int(ia),c_int(c),c_int(im))

    def read_annotation(self):
        '''Read the display annotation.'''
        lib.HP_3582A_read_annotation.restype = c_int
        arrmem = create_string_buffer("",200)
        iret = lib.HP_3582A_read_annotation(self.obj,arrmem,200)
        if( iret > 0 ):
            return arrmem.value
        else:
            return "None."

    def parse_channel_annotation(self,ann,c):
        '''Return the annotation for channel c.
        ann is the result of read_annotation().
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2' '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        if( ic == 1 ):
            annpart = ann[0:32]
        elif( ic == 2 ):
            annpart = ann[32:64]
        else:
            return( 'Bad Channel.' )
        return( ' '.join(annpart.split()) )

    def parse_average_annotation(self,ann):
        '''Return information on averaging from annotation ann.'''
        annpart = ann[96:118]
        return( ' '.join(annpart.split()) )

    def parse_bandwidth_annotation(self,ann):
        '''Return bandwidth information from annotation ann.'''
        annpart = ann[118:128]
        return( ' '.join(annpart.split()) )

    def get_spectrum(self):
        '''Read a spectrum. This returns a tuple: (frq,cha,chb,samps,aok,bok).
        frq is a NumPy array of frequencies.
        cha is a NumPy array of channel A amplitudes at the frq frequencies.
        chb is a NumPy array of channel B amplitudes at the frq frequencies.
        samps in the number of values in the arrays.
        aok is True if channel A was measured.
        bok is True if channel B was measured. '''
        freqs = (c_double*512)()
        chamem = (c_double*512)()
        chbmem = (c_double*512)()
        nsamp = c_int(0)
        a_avail = c_int(0)
        b_avail = c_int(0)
        iret = lib.HP_3582A_get_spectrum(self.obj,byref(freqs),byref(chamem),byref(chbmem),
                                         byref(nsamp),byref(a_avail),byref(b_avail))
        cha_d = fromiter( chamem, dtype=float, count=512 )
        chb_d = fromiter( chbmem, dtype=float, count=512 )
        return (freqs, cha_d, chb_d, nsamp.value, a_avail.value, b_avail.value)


class TEK_710A(GPIBInst):
    '''Sony/Tektronix RTD-710A Transient Digitizer Object.'''
    def __init__(self,a,g):
        super(TEK_710A,self).__init__()
        self.obj = lib.TEK_RTD710A_new( a, g )
        super(TEK_710A,self).set_obj(self.obj)

    def range(self,c,v):
        '''Set the sensitivity of channel c to v volts.
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2' '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        return lib.TEK_RTD710A_range(self.obj, c_int(ic), c_double(v))

    def offset(self,c,v):
        '''Set the offset of channel c to v volts.
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2' '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        return lib.TEK_RTD710A_offset(self.obj, c_int(ic), c_double(v))

    def coupling(self,c,m):
        '''Set the coupling mode for channel c.
        c may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2'
        m may be: 'ac', 'dc', 'gnd', 'ac_lf_reject', 'ac_hf_reject' '''
        try:
            ic = chancodes(c)
            im = couplecodes(m)
        except Exception, err:
            argerr()
        return lib.TEK_RTD710A_coupling(self.obj, c_int(ic), c_int(im))

    def bw_limit(self,b):
        '''Turn on bandwidth limiting (b=1) or turn it off (b=0).'''
        return lib.TEK_RTD710A_bw_limit(self.obj, c_int(b))

    def ch1_only(self,b):
        '''Turn on Channel 1 only mode (b=1) or turn it off (b=0).'''
        return lib.TEK_RTD710A_ch1_only(self.obj, c_int(b))

    def sample_interval(self,t):
        '''Set the interval between samples to be t seconds.'''
        return lib.TEK_RTD710A_sample_interval(self.obj, c_double(t))

    def samples(self,n):
        '''Set the number of samples to take to n.'''
        return lib.TEK_RTD710A_samples(self.obj, c_int(n))

    def sample_hold(self,b):
        '''Stop sampling (b=1) or start again (b=0).'''
        return lib.TEK_RTD710A_sample_hold(self.obj, c_int(b))

    def record_wave(self,waitms):
        '''Record a waveform. Allow waitms milliseconds for this to complete.'''
        return lib.TEK_RTD710A_record_wave(self.obj, c_int(waitms))

    def read_wave(self,location,c):
        '''Read a waveform back from a specified location number and channel, c.
        chan may be: 'a', 'cha', 'b', 'chb', 'ch1', 'ch2' '''
        try:
            ic = chancodes(c)
        except Exception, err:
            argerr()
        wave = (c_double*300000)()
        npoints = c_int(0)
        tstep = c_double(0.0)
        iret = lib.TEK_RTD710A_read_wave(self.obj, c_int(location), c_int(ic), byref(wave), 
                                      byref(npoints), byref(tstep))
        wave_d = fromiter( wave, dtype=float, count=npoints.value )
        return (wave_d, npoints.value, tstep.value)


if __name__ == '__main__':
    address = { 'HP_3582A':12, 'HP_3325A':6, 'HP_3457A':2, 'HP_8903E':14, 'TEK_710A':10 }

    if( False ):
        d = TEK_710A( address['TEK_710A'], 0 )
        d.range( 'ch1', 2 )
        d.sample_interval( eu(200,'ns') )
        d.record_wave( 1000 )
        (wave, nsamps, timestep) = d.read_wave( 1, 'ch1' )
        print nsamps, timestep
        plt.plot(wave[0:nsamps-1])
        plt.show()

    if( True ):
        m = HP_3457A( address['HP_3457A'], 0 )
        m.set_print_gpib(1)
        m.set_dump_gpib(1)
        m.set_function( 'freq', 'auto' )
        print m.read()

    if( False ):
        # Generate a 1kHz, 1V square wave.
        f = HP_3325A( address['HP_3325A'], 0 )
        f.set_print_gpib(0)
        f.set_print_error(0)
        f.set_function( 'sine' )
        f.set_frequency( 1, 'kHz' )
        f.set_amplitude( 1, 'v_rms' )

        # Get the signal's spectrum
        s = HP_3582A( address['HP_3582A'], 0 )
        s.set_print_gpib(0)
        s.set_print_error(0)
        s.set_input_mode( 'cha' )
        s.set_sensitivity( 'cha', 3 )
        s.start_measurement( 'rms', 5, 1 )
        (frq,cha,chb,samps,aok,bok) = s.get_spectrum()

        # Plot it.
        ann = s.read_annotation()
        ch1ann = s.parse_channel_annotation(ann,'cha')
        ch2ann = s.parse_channel_annotation(ann,'chb')
        avann = s.parse_average_annotation(ann)
        bwann = s.parse_bandwidth_annotation(ann)
        plt.plot(frq[0:samps-1],cha[0:samps-1])
        plt.title( 'HP 3582A Spectrum Analyzer' )
        plt.xlabel( 'Frequency (Hz)' )
        plt.ylabel( 'Amplitude (dBV)' )
        plt.grid( True )
        ax = plt.gca()
        plt.text( 0.6, 0.90, ch1ann, transform=ax.transAxes )
        plt.text( 0.6, 0.85, ch2ann, transform=ax.transAxes )
        plt.text( 0.6, 0.80, avann, transform=ax.transAxes )
        plt.text( 0.6, 0.75, bwann, transform=ax.transAxes )
        plt.show()

    
    #m = HP_3457A( 2, 0 )

    #f = HP_3325A( 6, 0 )
    ##f.string_test()
    #a# = f.array_test()
    #print a[1]
    #plt.plot(a)
    #plt.show()
    #f.set_print_gpib(1)
    #f.set_print_error(1)
    #f.set_frequency( 12345, 'Hz' )
    #f.set_amplitude( 800, 'mv' )
    #f.set_sweep_start_frequency(100,'Hz')
    #f.set_sweep_end_frequency(20,'kHz')
    #f#.set_sweep_time(5)
    #f.start_sweep_single(1)
    #f.set_frequency( 12345, 'Hz' )
    #f.set_amplitude( 500, 'mv' )

    #m.set_function('acv','auto')
    #print m.read()

    #f = HP_3457A( 2, 0 )
    #f.reset_bus()
    #f.set_print_gpib(1)
    #f.set_print_error(1)
    #f.check_open()
    #f.set_fast_mode(0)
    #f.set_function('acv','auto')
    #print f.read()
    #f.clear_fast_mode()
    #f.set_function('acv','auto')
    #print f.read()
    #print 'start wait'
    #f.wait_for_srq()
    #print 'end wait'
