// GPIBlib.cpp - Small library to control GPIB instruments.
// ========================================================
//
// Nick Glazzard  December 2005 Visa version on Windows.
//                December 2012 Linux GPIB version.
// -----------------------------------------------------

#define BUILDING_LIBRARY
#include "GPIBlib.h"

//====================================
// GpibInstrument Base Class methods.
//====================================

GpibInstrument::GpibInstrument( int address, int gpibif )
//-------------------------------------------------------
// Base instrument constructor.
{
  ENTER;
  _haveerror = false;
  _dispstatus = false;
  _address = -1;
  _gpibif = -1;

  // Allocate space for a message buffer. This is just assumed to work! Ugh.
  _message = (char*)malloc( GPIB_MSL ); 
  memset( _message, 0, GPIB_MSL );
  
  // Allocate space for GPIB input and output.
  _gpibout = (char*)malloc( GPIB_OL );
  _gpibin = (char*)malloc( GPIB_IL );
  if( _gpibout == NULL || _gpibin == NULL ){
    LocalSetError( "GpibInstrument constructor initialization failed!\n" );
  }
  
  // Initialize.
  else{
    memset( _namestring, 0, GPIB_NL );
    memset( _errstring, 0, GPIB_EL );
    memset( _gpibout, 0, GPIB_OL );
    memset( _gpibin, 0, GPIB_IL );
    _printerror = true; // Try to print error info during the initial open.
    _printgpib = false;
    _dumpgpibin = false;
    
    // Set a Visa format name string. No longer used for anything.
    sprintf( _namestring, "GPIB%d::%d::INSTR", gpibif, address );
    
    // If nothing opened yet, get a session handle to the Default Resource Manager.
    // This is left over from the Visa version, but may prove useful one day.
    if( _global_open_count == 0 ){
      _err = 0;
      _global_open_count = 1;
    }
    
    // Set the Default Resource Manager session handle in this object.
    _rsm = 1;
    _printerror = false; // Don't print errors anymore unless turned on explicitly.
    _address = address;
    _gpibif = gpibif;
    _termcharen = false;
    _termchar = 0x0a;  // Default termination character is LF.
  }
  LEAVE;
}

GpibInstrument::~GpibInstrument()
//-------------------------------
// Base Instrument destructor.
{
  ENTER;
  bool abort = ( _gpibout == NULL || _gpibin == NULL );
  if( abort ){
    if( _gpibout != NULL )
      free( _gpibout );
    if( _gpibin == NULL )
      free( _gpibin );
    _global_open_count = 0;
  }

  // If we are down to only 1 thing open, that must be the session handle.
  // Close it, and say we have nothing left open.
  // This is left over from the Visa version, but may prove useful one day.
  if( _global_open_count == 1 ){
    _global_open_count = 0;
    free( _gpibout );
    free( _gpibin );
  }
  LEAVE;
}

int GpibInstrument::DevOpen( int brd, int dev, GpibHandle* sess )
//---------------------------------------------------------------
// Open a device.
{
  ENTER;
  int ret = 0;
  *sess = ibdev( brd, dev, 0, TNONE, 1, 0 );
  if( *sess == (-1) ){
    iberr = 9999;
    ret |= 0x8000;
  }
  ibclr( *sess );
  LEAVE;
  return( ret );
}

int GpibInstrument::DevClose( GpibHandle sess )
//---------------------------------------------
// Close a device.
{
  return( 0 );
}

void GpibInstrument::Notify( const char* fmt, ... )
//-------------------------------------------------
// Output a message via notify_out().
{
  va_list args;

  va_start( args, fmt );
  vsprintf( _message, fmt, args );
  va_end( args );
  notify_out();
}

void GpibInstrument::notify_out( void )
//-------------------------------------
// Output a message string.
{
  fprintf( stderr, "%s", _message );
}

int GpibInstrument::IsError( int status )
//---------------------------------------
// See if status represents an error condition.
// Optionally make readable representation of status.
{
  if( _dispstatus )
    DisplayStatus( status );
  return( ( status & 0x8000 ) != 0 );
}

void GpibInstrument::ResetBus( void )
//-----------------------------------
// Reset the whole bus. Possibly.
{
  ENTER;
  memset( _gpibout, 0, GPIB_OL );
  memset( _gpibin, 0, GPIB_IL );
  ibsic( _gpibif );
  LEAVE;
}

char* GpibInstrument::GetLastError( void )
//----------------------------------------
// Return NULL if no error.
// Return pointer to internal error string if there was.
// Reset the _haveerror flag.
{
  ENTER;
  if( _haveerror ){
    _haveerror = false;
    LEAVE;
    return( _errstring );
  }
  else{
    LEAVE;
    return( NULL );
  }
}

bool GpibInstrument::SetStdAttributes( GpibHandle inst, char termchar )
//--------------------------------------------------------------------
// Set some default attributes for [inst]
{
  ENTER;
  _termchar = termchar;
  _err = ibtmo( inst, TNONE );
  if( IsError( _err ) ){
    GpibSetError();
    LEAVE;
    return( false );
  }
  _err = ibeos( inst, BIN|REOS|_termchar ); // BIN is needed for 82357A at least.
  if( IsError( _err ) ){
    GpibSetError();
    LEAVE;
    return( false );
  }
  _termcharen = true;
  if( _printgpib )
    printf("--------> TERMCHAR set to: 0x%x\n", _termchar );
  LEAVE;
  return( true );
}

bool GpibInstrument::ClearGPIBin( void )
//--------------------------------------
// Fill the GPIB input buffer with zeros.
{
  memset( _gpibin, 0, GPIB_IL );
  return( true );
}

bool GpibInstrument::ResetGPIB( GpibHandle inst )
//----------------------------------------------
// Reset the GPIB link with [inst].
// *NAFF WARNING* : Not sure this is at all correct! :-)
// It *does* reset the "hold" condition on the Sony/Tek RTD-710A which otherwise makes
// everything "stick" if a transfer fails. But maybe it is just an illegal command that
// happens to have that side effect...
{
  sprintf( _gpibout, "*RST\n" );
  return( SendGPIB( inst ) );
}

bool GpibInstrument::ClearDevice( GpibHandle inst )
//------------------------------------------------
// send a clear command to [inst].
{
  ENTER;
  _err = ibclr( inst );
 if( IsError( _err ) ){
    GpibSetError();
    if( _printgpib )
      Notify( "-------->ERROR\n" );
    LEAVE;
    return( false );
  }
  if( _printgpib )
    Notify( "-------->OK\n" );
  LEAVE;
  return( true );
}


bool GpibInstrument::SendGPIB( GpibHandle inst )
//---------------------------------------------
// Send the contents of _gpibout to [inst].
{
  ENTER;
  if( _printgpib )
    Notify( "GPIB-OUT>\"%s\"\n", _gpibout );
  _err = ibwrt( inst, (void*)_gpibout, (long)strlen(_gpibout) );
  if( IsError( _err ) ){
    GpibSetError();
    if( _printgpib )
      Notify( "-------->ERROR\n" );
    return( false );
  }
  if( _printgpib )
    Notify( "-------->OK\n" );
  return( true );
}

bool GpibInstrument::GetGPIB( GpibHandle inst )
//--------------------------------------------
// Read a string up to EOS from instrument [inst]
{
  ENTER;
  if( _printgpib )
    Notify( " GPIB-IN>\n" );
  _err = ibrd( inst, (void*)_gpibin, GPIB_IL-1 );
  if( IsError( _err ) ){
    GpibSetError();
    if( _printgpib )
      Notify( "-------->ERROR\n" );
    LEAVE;
    return( false );
  }
  if( _termcharen ){
    for( int i=0; i<GPIB_IL; i++ ){
      if( _gpibin[i] == _termchar ){
	_gpibin[i] = 0;
	if( _printgpib )
	  Notify( "-------->TERMCHAR FOUND (%d bytes in)\n", i );
	break;
      }
    }
  }
  if( _printgpib )
    Notify( "-------->OK \"%s\"\n", _gpibin );
  if( _dumpgpibin )
    dump_gpibin( strlen(_gpibin) );
  LEAVE;
  return( true );
}

bool GpibInstrument::GetBinaryGPIB( GpibHandle inst, int maxbytes, int& nread )
//----------------------------------------------------------------------------
// Read pure binary data from instrument [inst].
// Read at most [maxbytes] bytes. 
// Return the number of bytes read in [nread].
{
  ENTER;
  int bytes = maxbytes;

  // Nothing read yet.
  nread = 0;
  
  // If there is a termination character set, turn it off.
  // Otherwise, the transfer *will* stop when this data value turns up ... :-)
  if( _termcharen  ){
    _err = ibeos( inst, 0 );
    if( IsError( _err ) ){
      GpibSetError();
      LEAVE;
      return( false );
    }
  }

  // Try reading a pure binary block of up to maxbytes bytes.
  // The number of bytes succesfully read will go back into bytes.
  if( _printgpib )
    Notify( " GPIB-BIN-IN>\n" );
  _err = ibrd( inst, (void*)_gpibin, (long)maxbytes );
  if( IsError( _err ) ){
    GpibSetError();
    if( _printgpib )
      Notify( "-------->ERROR\n" );
    LEAVE;
    return( false );
  }
  if( _printgpib )
    Notify( "-------->OK: Read %d bytes.\n", bytes );
  
  // If there was a termination character set, set it up again.
  if( _termcharen  ){
    _err =  ibeos( inst, BIN|REOS|_termchar );
    if( IsError( _err ) ){
      GpibSetError();
      LEAVE;
      return( false );
    }
  }

  // Return number of bytes we got.
  nread = ibcnt;

  // Optionally, dump representation of the data.
  if( _dumpgpibin )
    dump_gpibin( nread );

  LEAVE;
  return( true );
}

void GpibInstrument::DisplayStatus( int status )
//----------------------------------------------
// Output a readable representation of a status word.
{
  char stbo[16*5+1];
  const char* dec[] = {"DCAS","DTAS","LACS","TACS", "ATN", "CIC", "REM", "LOK",
		       "CMPL","EVNT","SPOL", "RQS","SRQI", "END","TIMO", "ERR"};
  memset( stbo, 0, 16*5+1 );
  for( int i=0; i<16; i++ ){
    int m = 1 << i;
    if( status & m ){
      strcat( stbo, dec[i] );
      strcat( stbo, "," );
    }
  }
  Notify( "Status=[%s]\n", stbo );
}

bool GpibInstrument::SetInterfaceTimeOut( float ms )
//--------------------------------------------------
// Set the timeout is milliseconds for the interface card.
{
  int tcode = 0;
  ms *= 0.001f;
  if( ms < 1e-5 )
    // if 0-ish, wait forever.
    tcode = 0;
  else{
    tcode = (int)( 2 * log10f(ms/0.00001f) + 0.4f ) + 1;
    if( tcode > 17 )
      tcode = 17;
  }
  _err = ibtmo( _gpibif, tcode );
  if( IsError( _err ) )
    return( false );
  return( true );
}

bool GpibInstrument::WaitForSRQ( float msecs )
//--------------------------------------------
// Wait for some instrument to assert SRQ. 
{
  ENTER;
  if( ! SetInterfaceTimeOut( msecs ) )
    return( false );
  _err = ibwait( _gpibif, SRQI|TIMO );
  if( IsError( _err ) )
    return( false );
  if( ! SetInterfaceTimeOut( 0.0f ) )
    return( false );
  LEAVE;
  return( true );
}

bool GpibInstrument::ReadSTB( GpibHandle inst, char* status )
//-----------------------------------------------------------
// Read a status byte from [inst] by a serial poll. Usually after WaitForSRQ().
{
  ENTER;
  _err = ibrsp( inst, status );
  if( IsError( _err ) )
    return( false );
  LEAVE;
  if( _printgpib )
    Notify( "-------->OK: Read STatus Byte: %d.\n", (int)(*status) );
  return( true );
}

bool GpibInstrument::QueryGPIB( GpibHandle inst, const char* query )
//-----------------------------------------------------------------
// Send query to instrument [inst] and try to get a string back into _gpibin.
{
  ENTER;
  sprintf( _gpibout, "%s\n", query );
  if( ! SendGPIB( inst ) )
    return( false );
  if( ! GetGPIB( inst ) )
    return( false );
  return( true );
}

void GpibInstrument::GpibSetError( void )
//---------------------------------------
// Decode iberr to a string and set the error flag.
{
  char adinfo[40];

  ENTER;
  sprintf( adinfo, " Addr=%d, IF=%d", _address, _gpibif );
  switch( iberr ){
  case EDVR: sprintf( _errstring, "EDVR" ); break;
  case ECIC: sprintf( _errstring, "ECIC" ); break;
  case ENOL: sprintf( _errstring, "ENOL" ); break;
  case EADR: sprintf( _errstring, "EADR" ); break;
  case EARG: sprintf( _errstring, "EARG" ); break;
  case ESAC: sprintf( _errstring, "ESAC" ); break;
  case EABO: sprintf( _errstring, "EABO" ); break;
  case ENEB: sprintf( _errstring, "ENEB" ); break;
  case EDMA: sprintf( _errstring, "EDMA" ); break;
  case EOIP: sprintf( _errstring, "EOIP" ); break;
  case ECAP: sprintf( _errstring, "ECAP" ); break;
  case EFSO: sprintf( _errstring, "EFSO" ); break;
  case EBUS: sprintf( _errstring, "EBUS" ); break;
  case ESTB: sprintf( _errstring, "ESTB" ); break;
  case ESRQ: sprintf( _errstring, "ESRQ" ); break;
  case ETAB: sprintf( _errstring, "ETAB" ); break;
  case 9999: sprintf( _errstring, "DEVOPENFAILED" ); break;
  default: sprintf( _errstring, "UNKNOWN ERROR" ); break;
  }
  strncat( _errstring, adinfo, GPIB_EL-1 );
  _haveerror = true;
  if( _printerror ){
    Notify( "%s\n", _errstring );
  }
  LEAVE;
}

void GpibInstrument::LocalSetError( const char* errtext )
//-------------------------------------------------------
// Set the error string to [errtext] to notify a non-Gpib error.
// E.g. something out of range for a specific instrument.
// Set the error flag.
{
  char adinfo[40];

  sprintf( adinfo, " Addr=%d, IF=%d", _address, _gpibif );
  strncpy( _errstring, errtext, GPIB_EL-1 );
  strncat( _errstring, adinfo, GPIB_EL-1 );
  _haveerror = true;
  if( _printerror ){
    Notify( "%s\n", _errstring );
  }
}

void GpibInstrument::set_print_error( bool yes )
//----------------------------------------------
// Output error messages via Notify().
{
  _printerror = yes;
}

void GpibInstrument::set_print_gpib( bool yes )
//---------------------------------------------
// Output all traffic sent and received on the GPIB via Notify().
{
  _printgpib = yes;
}

void GpibInstrument::set_display_status( bool yes )
//-------------------------------------------------
// Output a readable representation of the current status bits after every GPIBlib call.
{
  _dispstatus = yes;
}

void GpibInstrument::set_dump_gpib( bool yes )
//-------------------------------------------------
// Output a repesentation of the binary data read in to _gpibin.
{
  _dumpgpibin = yes;
}

void GpibInstrument::dump_gpibin( int n )
//----------------------------------------
// Dump the contents of _gpibin, up to byte n, binary representation and character.
{
  for( int i=0; i<n; i++ ){
    if( i%10 == 0 )
      Notify( "\n" );
    Notify( "0x%x:%c,", _gpibin[i], _gpibin[i] );
  }
  Notify( "\n" );
}

// C bindings for Python Ctypes.

extern "C" {
  GpibInstrument* GpibInstrument_new( int a, int g )
  {return new GpibInstrument(a,g);}
  int GpibInstrument_ResetBus(GpibInstrument* s)
  {s->ResetBus(); return 1;}
  char* GpibInstrument_GetLastError(GpibInstrument* s)
  {return s->GetLastError();}
  int GpibInstrument_SetInterfaceTimeOut(GpibInstrument* s, float ms)
  {return(s->SetInterfaceTimeOut(ms)?1:0);}
  int GpibInstrument_set_print_error(GpibInstrument* s, bool yes)
  {s->set_print_error(yes?1:0); return 1;}
  int GpibInstrument_set_print_gpib(GpibInstrument* s, bool yes)
  {s->set_print_gpib(yes?1:0); return 1;}
  int GpibInstrument_set_display_status(GpibInstrument* s, bool yes)
  {s->set_display_status(yes?1:0); return 1;}
  int GpibInstrument_set_dump_gpib(GpibInstrument* s, bool yes)
  {s->set_dump_gpib(yes?1:0); return 1;}
}

#undef BUILDING_LIBRARY
