//==========================================
// HP 3325A Synthesiser/Function Generator.
//==========================================

#define BUILDING_LIBRARY
#include "include/GPIBlib.h"

Gpib_HP_3325A::Gpib_HP_3325A( int address, int gpibif ) : GpibInstrument( address, gpibif )
//-----------------------------------------------------------------------------------------
// Open a connection to an HP 3325A on GPIB address [address] (interface number [gpibif]).
{
  _hp3325a_opened = false;
  _waveform = -1;
  _freq = -1.0f;
  _amp = -1.0f;

  _err = DevOpen( gpibif, address, &_hp3325a );
  if( IsError( _err ) ){
    GpibSetError();
    return;
  }

  _global_open_count++;
  _hp3325a_opened = true;

  SetStdAttributes( _hp3325a );

  _waveform = WAVE_SINE;
  _freq = 1000.0f;
  _amp = 0.001f;
  _dc = 0.0;
  _phase = 0.0;
  _sweep_start = 0.0;
  _sweep_end = 0.0;
  _sweep_marker = 0.0;
  _sweep_mode = SWEEP_LINEAR;
  _sweep_time = 0.0;
  _modulate = MOD_NONE;
  _instrument_err = 0;
}

Gpib_HP_3325A::~Gpib_HP_3325A()
//-----------------------------
// Close the connection to the HP 3325A.
{
  if( _hp3325a_opened && _global_open_count > 1 ){
    DevClose( _hp3325a );
    _hp3325a_opened = false;
    _global_open_count--;
  }
}

bool Gpib_HP_3325A::check_open( void )
//------------------------------------
// Return true if open worked, else false.
// If false, set the error text appropriately.
{
  if( ! _hp3325a_opened ){
    LocalSetError( "HP3325A: Instrument open failed." );
    return( false );
  }
  return( true );
}

bool Gpib_HP_3325A::check_error( void )
//-------------------------------------
// Read the error status from the 3325A.
// Set the _instrument_err code and set the error string.
{
  if( ! check_open() )
    return( false );

  sprintf( _gpibout, "IER\n" );
  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! GetGPIB( _hp3325a ) )
    return( false );
  if( sscanf( _gpibin, "ER%d", &_instrument_err ) != 1 )
    return( true ); //false );

  switch( _instrument_err ){
  case 0:
    return( true );
  case 1:
    LocalSetError( "HP3325A: Entry parameter out of bounds." );
    return( false );
  case 2:
    LocalSetError( "HP3325A: Invalid delimiter received." );
    return( false );
  case 3:
    LocalSetError( "HP3325A: Frequency too large for function." );
    return( false );
  case 4:
    LocalSetError( "HP3325A: Sweep time too small or too large." );
    return( false );
  case 5:
    LocalSetError( "HP3325A: Offset and Amplitude are incompatible with one another." );
    return( false );
  case 6:
    LocalSetError( "HP3325A: Sweep frequency error." );
    return( false );
  case 7:
    LocalSetError( "HP3325A: Unknown mnemonic received." );
    return( false );
  case 8:
    LocalSetError( "HP3325A: Unrecognizable data character received." );
    return( false );
  case 9:
    LocalSetError( "HP3325A: Option required is not installed." );
    return( false );
  default:
    LocalSetError( "HP3325A: Unknown error returned from 3325A." );
    return( false );
  }
}

bool Gpib_HP_3325A::set_function( int waveform )
//----------------------------------------------
// Specify the waveform we want.
{
  if( ! check_open() )
    return( false );

  switch( waveform ){
  case WAVE_DC:
  case WAVE_SINE:
  case WAVE_SQUARE:
  case WAVE_TRIANGLE:
  case WAVE_RAMP_UP:
  case WAVE_RAMP_DOWN:
    sprintf( _gpibout, "FU%d\n", waveform );
    if( ! SendGPIB( _hp3325a ) )
      return( false );
    break;
  default:
    LocalSetError( "HP3325A: Invalid waveform requested." );
    return( false );
  }

  if( ! check_error() )
    return( false );

  _waveform = waveform;
  return( true );
}

bool Gpib_HP_3325A::set_frequency( float freq, int units )
//--------------------------------------------------------
// Set the frequency we want.
{
  if( ! check_open() )
    return( false );

  switch( units ){
  case FREQ_KHZ:
    freq *= 1000.0f;
    break;
  case FREQ_MHZ:
    freq *= 1000000.0f;
    break;
  case FREQ_GHZ:
    freq *= 1000000000.0f;
    break;
  }

  if( freq < 100000.0f )
    sprintf( _gpibout, "FR%12.6fHZ\n", freq );
  else
    sprintf( _gpibout, "FR%12.3fHZ\n", freq );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _freq = freq;
  return( true );
}

bool Gpib_HP_3325A::set_amplitude( float amp, int units )
//-------------------------------------------------------
// Set the amplitude we want.
{
  bool rms = false;
  
  if( ! check_open() )
    return( false );
  
  if( units == AMP_MV || units == AMP_MV_PP || units == AMP_MV_RMS )
    amp *= 0.001f;
  rms = ( units == AMP_V_RMS || units == AMP_MV_RMS );
  
  if( units == AMP_DBM ){
    sprintf( _gpibout, "AM%.1fDB\n", amp );
    
  }
  
  else{
    if( amp < 0.0 ){
      LocalSetError( "HP3325A: Invalid amplitude requested." );
      return( false );
    }
    else{
      if( amp >= 10.0f ) // 10.00 is max allowed.
        sprintf( _gpibout, "AM%.2f%s\n", amp, ( rms ? "VR" : "VO" ) );
      else if( amp >= 1.0f )
        sprintf( _gpibout, "AM%.3f%s\n", amp, ( rms ? "VR" : "VO" ) );
      else
        sprintf( _gpibout, "AM%d%s\n", (int)( amp * 1000.0f + 0.5f ), ( rms ? "MR" : "MV" ) );
    }
  }
  
  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );
  
  if( units == AMP_DBM ){
    _amp = 0.2236f * powf( 10.0f, ( amp / 20.0f ) ); // Convert dBm(50 ohm) to volts p-p.
  }
  else{
    if( rms )
      _amp = 2.828427f * amp; // Convert RMS(sine) to p-p.
    else
      _amp = amp;
  }
  
  return( true );
}

bool Gpib_HP_3325A::set_dc_offset( float dc, int units )
//------------------------------------------------------
// Set the DC offset level.
{
  if( ! check_open() )
    return( false );
  
  if( units == AMP_MV )
    dc *= 0.001f;
  
  if( dc >= 1.0f )
    sprintf( _gpibout, "DC%.3fVO\n", dc );
  else
    sprintf( _gpibout, "DC%dMV\n", (int)( dc * 1000.0f + 0.5f ) );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _dc = dc;
  
  return( true );
}

bool Gpib_HP_3325A::set_phase( float degrees )
//--------------------------------------------
// Set the phase offset w.r.t. arbitrary starting phase or
// assigned zero phase.
{
  if( ! check_open() )
    return( false );

  if( degrees > 719 )degrees = 719; else if( degrees < -719 )degrees = -719;

  sprintf( _gpibout, "PH%dDE\n", (int)( degrees + 0.5f ) );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _phase = degrees;

  return( true );
}

bool Gpib_HP_3325A::set_sweep_start_frequency( float freq, int units )
//--------------------------------------------------------------------
// Set the sweep start frequency.
{
  if( ! check_open() )
    return( false );
  
  switch( units ){
  case FREQ_KHZ:
    freq *= 1000.0f;
    break;
  case FREQ_MHZ:
    freq *= 1000000.0f;
    break;
  case FREQ_GHZ:
    freq *= 1000000000.0f;
    break;
  }

  if( freq < 100000.0f )
    sprintf( _gpibout, "ST%12.6fHZ\n", freq );
  else
    sprintf( _gpibout, "ST%12.3fHZ\n", freq );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _sweep_start = freq;

  return( true );
}

bool Gpib_HP_3325A::set_sweep_end_frequency( float freq, int units )
//------------------------------------------------------------------
// Set the sweep end frequency.
{
  if( ! check_open() )
    return( false );

  switch( units ){
  case FREQ_KHZ:
    freq *= 1000.0f;
    break;
  case FREQ_MHZ:
    freq *= 1000000.0f;
    break;
  case FREQ_GHZ:
    freq *= 1000000000.0f;
    break;
  }

  if( freq < 100000.0f )
    sprintf( _gpibout, "SP%12.6fHZ\n", freq );
  else
    sprintf( _gpibout, "SP%12.3fHZ\n", freq );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _sweep_end = freq;

  return( true );
}

bool Gpib_HP_3325A::set_sweep_marker_frequency( float freq, int units )
//---------------------------------------------------------------------
// Set the sweep marker frequency.
{
  if( ! check_open() )
    return( false );

  switch( units ){
  case FREQ_KHZ:
    freq *= 1000.0f;
    break;
  case FREQ_MHZ:
    freq *= 1000000.0f;
    break;
  case FREQ_GHZ:
    freq *= 1000000000.0f;
    break;
  }

  if( freq < 100000.0f )
    sprintf( _gpibout, "MF%12.6fHZ\n", freq );
  else
    sprintf( _gpibout, "MF%12.3fHZ\n", freq );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _sweep_marker = freq;

  return( true );
}

bool Gpib_HP_3325A::set_sweep_mode( int sweep_mode )
//--------------------------------------------------
// Set the sweep mode to linear (SWEEP_LINEAR) or log (SWEEP_LOG).
{
  if( ! check_open() )
    return( false );

  if( sweep_mode != SWEEP_LINEAR && sweep_mode != SWEEP_LOG ){
    LocalSetError( "HP3325A: Invalid sweep mode requested." );
    return( false );
  }

  sprintf( _gpibout, "SM%d\n", sweep_mode );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _sweep_mode = sweep_mode;

  return( true );
}

bool Gpib_HP_3325A::set_sweep_time( float secs )
//----------------------------------------------
// Set the time to take for a sweep.
{
  if( ! check_open() )
    return( false );

  if( secs < 0.01f || secs > 99.99f ){
    LocalSetError( "HP3325A: Invalid sweep time requested." );
    return( false );
  }

  sprintf( _gpibout, "TI%.2fSE\n", secs );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  _sweep_time = secs;

  return( true );
}

bool Gpib_HP_3325A::start_sweep_single( bool waitfin )
//----------------------------------------------------
// Do a single sweep from start freq to end freq in sweep_time secs.
// If [waitfin], wait for it to finish.
{
  if( ! check_open() )
    return( false );
  
  // If wait: Enable the sweep stop event. Trigger SRQ on stop.
  if( waitfin ){
    sprintf( _gpibout, "MSB\n" );
    if( ! SendGPIB( _hp3325a ) )
      return( false );
    if( ! check_error() )
      return( false );
  }

  // Start the sweep.
  sprintf( _gpibout, "SSSS\n" );
  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );
  
  // If wait: Start waiting for a stop event.
  if( waitfin ){
    if( ! WaitForSRQ( 1000.0f*(_sweep_time+2.0f) ) )
      return( false );
  }
  
  return( true );
}

bool Gpib_HP_3325A::start_sweep_continuous( void )
//------------------------------------------------
// Start doing continuous sweeps.
{
  if( ! check_open() )
    return( false );

  sprintf( _gpibout, "SC\n" );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  return( true );
}

bool Gpib_HP_3325A::use_amplitude_modulation( bool yes )
//------------------------------------------------------
// Turn amplitude modulation on or off.
{
  if( ! check_open() )
    return( false );

  sprintf( _gpibout, "MA%d\n", ( yes ? 1 : 0 ) );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  if( yes )
    _modulate = MOD_AMP;
  else
    _modulate = MOD_NONE;

  return( true );
}

bool Gpib_HP_3325A::use_phase_modulation( bool yes )
//--------------------------------------------------
// Turn phase modulation on or off.
{
  if( ! check_open() )
    return( false );


  sprintf( _gpibout, "MP%d\n", ( yes ? 1 : 0 ) );

  if( ! SendGPIB( _hp3325a ) )
    return( false );
  if( ! check_error() )
    return( false );

  if( yes )
    _modulate = MOD_PHASE;
  else
    _modulate = MOD_NONE;

  return( true );
}

extern "C"{
  Gpib_HP_3325A* HP_3325A_new( int a, int g )
  { return new Gpib_HP_3325A(a,g); }
  int HP_3325A_check_open( Gpib_HP_3325A* s )
  { return(s->check_open()?1:0); }
  int HP_3325A_check_error( Gpib_HP_3325A* s )
  { return(s->check_error()?1:0); }

  int HP_3325A_set_function( Gpib_HP_3325A* s, int waveform )
  { return(s->set_function(waveform)?1:0); }
  int HP_3325A_set_frequency( Gpib_HP_3325A* s, float freq, int units )
  { return(s->set_frequency(freq, units)?1:0); }
  int HP_3325A_set_amplitude( Gpib_HP_3325A* s, float amp, int units )
  { return(s->set_amplitude(amp, units)?1:0); }
  int HP_3325A_set_dc_offset( Gpib_HP_3325A* s, float dc, int units )
  { return(s->set_dc_offset(dc, units)?1:0); }
  int HP_3325A_set_phase( Gpib_HP_3325A* s, float degrees )
  { return(s->set_phase(degrees)?1:0); }

  int HP_3325A_set_sweep_start_frequency( Gpib_HP_3325A* s, float freq, int units )
  { return(s->set_sweep_start_frequency(freq, units)?1:0); }
  int HP_3325A_set_sweep_end_frequency( Gpib_HP_3325A* s, float freq, int units )
  { return(s->set_sweep_end_frequency(freq, units)?1:0); }
  int HP_3325A_set_sweep_marker_frequency( Gpib_HP_3325A* s, float freq, int units )
  { return(s->set_sweep_marker_frequency(freq, units)?1:0); }
  int HP_3325A_set_sweep_mode( Gpib_HP_3325A* s, int sweep_mode )
  { return(s->set_sweep_mode(sweep_mode)?1:0); }
  int HP_3325A_set_sweep_time( Gpib_HP_3325A* s, float secs )
  { return(s->set_sweep_time(secs)?1:0); }

  int HP_3325A_start_sweep_single( Gpib_HP_3325A* s, int waitfin )
  { return(s->start_sweep_single((waitfin==1))?1:0); }

  int HP_3325A_start_sweep_continuous( Gpib_HP_3325A* s )
  { return(s->start_sweep_continuous()?1:0); }

  int HP_3325A_use_amplitude_modulation( Gpib_HP_3325A* s, int yes )
  { return(s->use_amplitude_modulation(yes)?1:0); }
  int HP_3325A_use_phase_modulation( Gpib_HP_3325A* s, int yes )
  { return(s->use_phase_modulation(yes)?1:0); }

  int test_arrs( double* darr, int* narr )
  { darr[0]=1.1; darr[1]=2.0; darr[2]=3.3; *narr = 3; return 4; }

  int string_test( char* cret, int max )
  { strncpy(cret,"Something or other.",max-1); return(strlen(cret)); }
}

#undef BUILDING_LIBRARY
