//=====================
// HP 3457A Multimeter.
//=====================

#define BUILDING_LIBRARY
#include "GPIBlib.h"

Gpib_HP_3457A::Gpib_HP_3457A( int address, int gpibif ) : GpibInstrument( address, gpibif )
//-----------------------------------------------------------------------------------------
// Open a connection to an HP 3457A on GPIB address [address] (interface number [gpibif]).
{
  _hp3457a_opened = false;

  _err = DevOpen( gpibif, address, &_hp3457a );
  if( IsError( _err ) ){
    GpibSetError();
    return;
  }

  _global_open_count++;
  _hp3457a_opened = true;

  _function = DCV;
  _instrument_err = 0;

  // We need to set the correct terminating character for this beast.
  SetStdAttributes( _hp3457a );

  sprintf( _gpibout, "PRESET\n" );
  if( ! SendGPIB( _hp3457a ) )
    _instrument_err = 999;
}

Gpib_HP_3457A::~Gpib_HP_3457A()
//-----------------------------
// Close the connection to the HP 3457A.
{
  if( _hp3457a_opened && _global_open_count > 1 ){
    sprintf( _gpibout, "RESET\n" );
    SendGPIB( _hp3457a );
    DevClose( _hp3457a );
    _hp3457a_opened = false;
    _global_open_count--;
  }
}

bool Gpib_HP_3457A::check_open( void )
//------------------------------------
// Return true if open worked, else false.
// If false, set the error text appropriately.
{
  if( ! _hp3457a_opened ){
    LocalSetError( "HP3457A: Instrument open failed." );
    return( false );
  }
  return( true );
}

bool Gpib_HP_3457A::set_function( int function, double vmax )
//-----------------------------------------------------------
// Set the function to perform and the range to use. The range
// is function dependent.
//
// [function] should be a FUNCTION CODE defined in GPIBlib.h.
// [vmax] should be the maximum value (volts, amps or ohms) that one 
// wishes to measure.
//
// DCV    ACV    ACDCV  OHM    OHMF   DCI    ACI    ACDCI  FREQ   PERIOD
// ---------------------------------------------------------------------
// AUTO   AUTO   AUTO   AUTO   AUTO   AUTO   AUTO   AUTO   AUTO   AUTO
//  30m    30m    30m     30     30   300u                   
// 300m   300m   300m    300    300     3m
//   3V     3V     3V     3K     3K    30m    30m    30m
//  30V    30V    30V    30K    30K   300m   300m   300m
// 300V   300V   300V   300K   300K    1A5     1A     1A
//                        3M     3M
//                       30M    30M
//                        3G     3G
{
  if( ! _hp3457a_opened ){
    LocalSetError( "HP3457A: Instrument open failed." );
    return( false );
  }
  if( function < DCV || function > PERIOD ){
    LocalSetError( "HP3457A: Invalid function code." );
    return( false );
  }
  switch( function ){
  case DCV:
  case ACV:
  case ACDCV:
    if( vmax > 300.0 )
      vmax = 300.0;
    break;
  case ACI:
  case ACDCI:
    if( vmax > 1.0 )
      vmax = 1.0;
    break;
  case DCI:
    if( vmax > 1.5 )
      vmax = 1.5;
    break;
  case FREQ:
  case PERIOD:
    vmax = -1.0; // There are other possibilities, but ... :-)
    break;
  case OHM:
  case OHMF:
    if( vmax > 3e9 )
      vmax = 3e9;
    break;
  }

  // Stop triggering.
  sprintf( _gpibout, "TRIG HOLD\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );

  // Set the function and range desired.
  if( vmax < 0.0 )
    sprintf( _gpibout, "FUNC %d,AUTO\n", function );
  else
    sprintf( _gpibout, "FUNC %d,%f\n", function, vmax );
  if( ! SendGPIB( _hp3457a ) )
    return( false );

  // Trigger when data output buffer empty.
  // I.e. after the host has read the data value.
  sprintf( _gpibout, "TRIG SYN\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );

  _function = function;
  return( true );
}

bool Gpib_HP_3457A::set_fast_mode( bool acfast )
//----------------------------------------------
// Read as quickly as possible at 3.5 digit resolution.
// If [acfast] turn on fast AC measurements. This should really only be
// done for frequencies over 400Hz.
{
  if( ! _hp3457a_opened ){
    LocalSetError( "HP3457A: Instrument open failed." );
    return( false );
  }

  sprintf( _gpibout, "DELAY\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );
  sprintf( _gpibout, "AZERO 0\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );
  sprintf( _gpibout, "NPLC 0\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );
  sprintf( _gpibout, "TIMER 0.1\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );
  if( acfast ){
    sprintf( _gpibout, "ACBAND 5000\n" );
    if( ! SendGPIB( _hp3457a ) )
      return( false );
  }
  return( true );
}

bool Gpib_HP_3457A::clear_fast_mode( void )
//-----------------------------------------
// Return to normal mode.
{
  if( ! ClearDevice( _hp3457a ) )
    return( false );
  sprintf( _gpibout, "PRESET\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );
  sprintf( _gpibout, "NPLC 10\n" );
  if( ! SendGPIB( _hp3457a ) )
    return( false );
  return( true );
}

bool Gpib_HP_3457A::read_simple( double* meas )
//---------------------------------------------
// Read a measurement.
{
  *meas = 0.0;
  if( ! _hp3457a_opened ){
    LocalSetError( "HP3457A: Instrument open failed." );
    return( false );
  }

  // Get a string from the 3457A's output buffer.
  if( ! GetGPIB( _hp3457a ) )
    return( false );

  // Convert it to a double with no questions asked!
  *meas = atof( _gpibin );
  return( true );
}

// C bindings for Python Ctypes.

extern "C" {
  Gpib_HP_3457A* HP_3457A_new( int a, int g )
  { return new Gpib_HP_3457A(a,g); }
  int HP_3457A_check_open( Gpib_HP_3457A* s )
  { return(s->check_open()?1:0); }
  int HP_3457A_set_function( Gpib_HP_3457A* s, int f, double v )
  { return(s->set_function(f,v)?1:0); }
  int HP_3457A_set_fast_mode( Gpib_HP_3457A* s, int a )
  { return(s->set_fast_mode(a==1)?1:0); }
  int HP_3457A_clear_fast_mode( Gpib_HP_3457A* s )
  { return(s->clear_fast_mode()?1:0); }
  double HP_3457A_read_simple( Gpib_HP_3457A* s )
  { double m; s->read_simple(&m); return(m); }
}

#undef BUILDING_LIBRARY
