//===========================
// HP 3582A Spectrum Analyzer
//===========================

#define BUILDING_LIBRARY
#include "GPIBlib.h"

Gpib_HP_3582A::Gpib_HP_3582A( int address, int gpibif ) : GpibInstrument( address, gpibif )
//-----------------------------------------------------------------------------------------
// Open a connection to an HP 3582A on GPIB address [address] (interface number [gpibif]).
{
  _hp3582a_opened = false;

  _err = DevOpen( gpibif, address, &_hp3582a );
  if( IsError( _err ) ){
    GpibSetError();
    return;
  }

  _global_open_count++;
  _hp3582a_opened = true;

  _instrument_err = 0;

  // We need to set the correct terminating character for this beast.
  SetStdAttributes( _hp3582a, 0x0d );

  // Send a command to set the "preset" state of the instrument.
  sprintf( _gpibout, "PRS\n" );
  if( ! SendGPIB( _hp3582a ) )
    _instrument_err = 999;

  // Set our frequency ranges and input and display states to match the
  // preset instrument state.
  _frlo = 0.0;
  _frhi = 25000.0;
  _disp_chan = CHA;
  _inp_chan = CHA;
  _ampref = 0.0;
}

Gpib_HP_3582A::~Gpib_HP_3582A()
//-----------------------------
// Close the connection to the HP 3582A.
{
  if( _hp3582a_opened && _global_open_count > 1 ){
    DevClose( _hp3582a );
    _hp3582a_opened = false;
    _global_open_count--;
  }
}

bool Gpib_HP_3582A::set_input_mode( int channels )
//------------------------------------------------
// Use CH A (CHA), CH B (CHB) or both (CHBOTH).
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  int code;
  switch( channels ){
  case CHA: code = 1; break;
  case CHB: code = 3; break;
  case CHBOTH: code = 2; break;
  default:
    LocalSetError( "HP3582A: Invalid input channel code." );
    return( false );
  }

  sprintf( _gpibout, "IM%d\n", code );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  _inp_chan = channels;

  return( true );
}

bool Gpib_HP_3582A::coupling( int channel, int mode )
//---------------------------------------------------
// Set the input [channel] (CHA, CHB) coupling to AC_COUPLED or DC_COUPLED
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  char* chcode;
  int modecode;

  switch( channel ){
  case CHA: chcode = (char*)"AC"; break;
  case CHB: chcode = (char*)"BC"; break;
  default:
    LocalSetError( "HP3582A: Invalid input channel code." );
    return( false );
  }

  switch( mode ){
  case AC_COUPLED: modecode = 1; break;
  case DC_COUPLED: modecode = 2; break;
  default:
    LocalSetError( "HP3582A: Invalid coupling mode code." );
    return( false );
  }

  sprintf( _gpibout, "%s%d\n", chcode, modecode );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  return( true );
}

bool Gpib_HP_3582A::set_sensitivity( int channel, double volts )
//--------------------------------------------------------------
// Set the input [channel] (CHA, CHB) sensitivity.
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  int scode;
  if( volts <= 3 MV )
    scode = 10;
  else if( volts <= 10 MV )
    scode = 9;
  else if( volts <= 30 MV )
    scode = 8;
  else if( volts <= 100 MV )
    scode = 7;
  else if( volts <= 300 MV )
    scode = 6;
  else if( volts <= 1.0 )
    scode = 5;
  else if( volts <= 3.0 )
    scode = 4;
  else if( volts <= 10.0 )
    scode = 3;
  else if( volts <= 30.0 )
    scode = 2;
  else{
    LocalSetError( "HP3582A: Invalid sensitivity (voltage too high)." );
    return( false );
  }

  char* chcode;
  switch( channel ){
  case CHA: chcode = (char*)"AS"; break;
  case CHB: chcode = (char*)"BS"; break;
  default:
    LocalSetError( "HP3582A: Invalid input channel code." );
    return( false );
  }

  sprintf( _gpibout, "%s%d\n", chcode, scode );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  return( true );
}

bool Gpib_HP_3582A::set_sensitivity_db( int channel, double fsdb )
//----------------------------------------------------------------
// Set the input [channel] (CHA, CHB) sensitivity.
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  int scode;
  if( fsdb <= -50.0 )
    scode = 10;
  else if( fsdb <= -40.0 )
    scode = 9;
  else if( fsdb <= -30.0 )
    scode = 8;
  else if( fsdb <= -20.0 )
    scode = 7;
  else if( fsdb <= -10.0 )
    scode = 6;
  else if( fsdb <= 0.0 )
    scode = 5;
  else if( fsdb <= 10.0 )
    scode = 4;
  else if( fsdb <= 20.0 )
    scode = 3;
  else if( fsdb <= 30.0 )
    scode = 2;
  else{
    LocalSetError( "HP3582A: Invalid sensitivity (voltage too high)." );
    return( false );
  }

  char* chcode;
  switch( channel ){
  case CHA: chcode = (char*)"AS"; break;
  case CHB: chcode = (char*)"BS"; break;
  default:
    LocalSetError( "HP3582A: Invalid input channel code." );
    return( false );
  }

  sprintf( _gpibout, "%s%d\n", chcode, scode );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  return( true );
}

bool Gpib_HP_3582A::set_frequency( int mode, double span, double freq )
//---------------------------------------------------------------------
// Set the frequency range. Mode may be:
// FULL_SPAN     : 0->25kHz [span] and [freq] ignored.
// ZERO_START    : 0->[span], [freq] ignored.
// SET_START     : [freq]->[freq]+[span]
// SET_CENTER    : [freq]-[span]/2 -> [freq]+[span]/2
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  if( span < 1.0 || span > 25000.0 ){
    LocalSetError( "HP3582A: Invalid frequency span." );
    return( false );
  }
  if( freq < 0.0 || freq > 25000.0 ){
    LocalSetError( "HP3582A: Invalid frequency specified." );
    return( false );
  }

  switch( mode ){
  case FULL_SPAN:
    _frlo = 0.0;
    _frhi = 25000.0;
    sprintf( _gpibout, "MD1\n" );
    break;
  case ZERO_START:
    _frlo = 0.0;
    _frhi = span;
    sprintf( _gpibout, "MD2SP%d\n", freq_to_span( span ) );
    break;
  case SET_START:
    _frlo = freq;
    _frhi = freq + span;
    if( _frhi > 25000.0 ){
      LocalSetError( "HP3582A: Invalid start+span specified. >25000Hz." );
      return( false );
    }
    sprintf( _gpibout, "MD3SP%dAD%d\n", freq_to_span( span ), (int)freq );
    break;
  case SET_CENTER:
    _frlo = freq - 0.5 * span;
    if( _frlo < 0.0 ){
      LocalSetError( "HP3582A: Invalid start-span specified. <0Hz." );
      return( false );
    }
    _frhi = freq + 0.5 * span;
    if( _frhi > 25000.0 ){
      LocalSetError( "HP3582A: Invalid start+span specified. >25000Hz." );
      return( false );
    }
    sprintf( _gpibout, "MD4SP%dAD%d\n", freq_to_span( span ), (int)freq );
    //sprintf( _gpibout, "MD4AD%dSP%d\n", (int)freq, freq_to_span( span ) );
    break;
  default:
    LocalSetError( "HP3582A: Invalid frequency range mode." );
    return( false );
  }
  if( ! SendGPIB( _hp3582a ) )
    return( false );
  return( true );
}

int Gpib_HP_3582A::freq_to_span( double span )
//--------------------------------------------
// Convert a frequency span to a span code.
{
  if( span <= 1.0 )return( 1 );
  if( span <= 2.5 )return( 2 );
  if( span <= 5.0 )return( 3 );
  if( span <= 10.0 )return( 4 );
  if( span <= 25.0 )return( 5 );
  if( span <= 50.0 )return( 6 );
  if( span <= 100.0 )return( 7 );
  if( span <= 250.0 )return( 8 );
  if( span <= 500.0 )return( 9 );
  if( span <= 1000.0 )return( 10 );
  if( span <= 2500.0 )return( 11 );
  if( span <= 5000.0 )return( 12 );
  if( span <= 10000.0 )return( 13 );
  if( span <= 25000.0 )return( 14 );
  return( 14 );
}

bool Gpib_HP_3582A::display( int channel, int ymode )
//---------------------------------------------------
// Select which channel to display and in which Y scale mode
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  // Turn off 0 or 1 channels. Turn on 1 or 2.
  // NASTY TRICK: offcode is used as an on code if offmode is 1.
  char* oncode;
  char* offcode;
  int offmode;
  switch( channel ){
  case CHA:    oncode = (char*)"AA"; offcode = (char*)"AB"; offmode = 0; break;
  case CHB:    oncode = (char*)"AB"; offcode = (char*)"AA"; offmode = 0; break;
  case CHBOTH: oncode = (char*)"AB"; offcode = (char*)"AA"; offmode = 1; break;
  default:
    LocalSetError( "HP3582A: Invalid display channel code." );
    return( false );
  }

  if( ymode < 1 || ymode > 3 ){
    LocalSetError( "HP3582A: Invalid Y mode." );
    return( false );
  }

  sprintf( _gpibout, "%s%d%s1SC%d\n", offcode, offmode, oncode, ymode );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  _disp_chan = channel;

  return( true );
}

bool Gpib_HP_3582A::set_amp_ref_level_db( double refdb )
//------------------------------------------------------
// Set the full scale level of the display to [refdb] dB below Full Scale.
// [refdb] should be a multiple of 10 and acts the same way whether +ve or -ve.
{
  if( refdb < 0.0 )refdb = -refdb;
  int irefdb = ( (int)refdb ) / 10;
  irefdb++;
  if( irefdb > 9 )irefdb = 9; // -80dB is lowest allowed.

  sprintf( _gpibout, "AM%d\n", irefdb );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  _ampref = ( -10.0 * ( irefdb - 1 ) );
  return( true );
}

double Gpib_HP_3582A::get_amp_ref_level_db( void )
//------------------------------------------------
// Return the (fixed up) amplitude reference level.
{
  return( _ampref );
}

bool Gpib_HP_3582A::window( int windowtype )
//------------------------------------------
// Set the passband shape to FLATTOP, HANNING or UNIFORM.
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  if( windowtype < FLATTOP || windowtype > UNIFORM ){
    LocalSetError( "HP3582A: Invalid window function." );
    return( false );
  }

  sprintf( _gpibout, "PS%d\n", windowtype );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  return( true );
}

bool Gpib_HP_3582A::start_measurement( int avtype, int count, int maxwait )
//-------------------------------------------------------------------------
// Start measuring (again) using averaging mode [avtype] (AVERAGE_OFF or AVERAGE_RMS for
// now), and doing [count] averages.
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }
  
  if( avtype < AVERAGE_OFF || avtype > AVERAGE_RMS ){
    LocalSetError( "HP3582A: Invalid averaging mode." );
    return( false );
  }

  // Check some basic setting compatibilities.
  if( _disp_chan == CHBOTH && _inp_chan != CHBOTH ){
    LocalSetError( "HP3582A: Display set to BOTH channels, but Input NOT set to BOTH channels." );
    return( false );
  }
  if( _inp_chan != CHBOTH ){
    if( _disp_chan != _inp_chan ){
    LocalSetError( "HP3582A: Input and Display channels differ and Input is not BOTH." );
    return( false );
    }
  }
  
  // Ensure averaging count is a power of 2. Convert to a count code.
  int v, s;
  for( v=1,s=0; v<count; ){
    v <<= 1;
    s++;
  }
  if( s < 2 )
    s = 2;
  s--;
  if( v > 32 )
    s -= 4;
  
  // Set averaging type and count as required.
  sprintf( _gpibout, "AV%d\n", avtype );
  if( ! SendGPIB( _hp3582a ) )
    return( false );
  
  if( avtype == AVERAGE_RMS ){
    if( v > 32 )
      sprintf( _gpibout, "SH1NU%d\n", s );
    else
      sprintf( _gpibout, "SH0NU%d\n", s );
    if( ! SendGPIB( _hp3582a ) )
      return( false );
  }
  
  // Send a reset average.
  sprintf( _gpibout, "RE\n" );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  // Begin a polling loop looking for the single sweep or average to end.
  int waitloop;
  unsigned char inb;
  
  for( waitloop=0; waitloop<maxwait; waitloop++ ){
    if( waitloop == 0 )
      sprintf( _gpibout, "LST0\n" );
    else
      sprintf( _gpibout, "LST1\n" );
    if( ! SendGPIB( _hp3582a ) )
      return( false );
    
    int nread;
    GetBinaryGPIB( _hp3582a, 1, nread );
    inb = _gpibin[0];
    
    // We want bit 5 to be set for single sweep or 6 for averages.
    if( waitloop > 1 ){
      if( avtype == AVERAGE_RMS ){
        if( ( inb & 64 ) != 0 )
          break;
      }
      else{
        if( ( inb & 16 ) != 0 )
          break;
      }
    }
    sleep( 5 );
  }

  if( waitloop == maxwait )
    return( false );

  return( true );
}

bool Gpib_HP_3582A::read_data( int& ndata, double* data, int maxdata )
//--------------------------------------------------------------------
// Read a spectrum (or time record) from the 3582A.
{
  char num[20];

  ndata = 0;
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  // Send List DiSplay command.
  sprintf( _gpibout, "LDS\n" );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  // Get whatever comes back.
  if( ! GetGPIB( _hp3582a ) )
    return( false );

  // Parse the returned string. Count the number of data items.
  char* ip;
  int i;
  
  _ndata = 0;
  for( ip=_gpibin,i=0; *ip; ip++ ){
    if( isdigit(*ip) || ( *ip == '+') || (*ip == '-') || (*ip == 'E' ) || (*ip == '.') ){
      if( i < 19 ){
        num[i] = *ip;
        ++i;
      }
    }
    else if( (*ip == ',') || (*ip == '\r') ){
        num[i] = 0;
        _data[_ndata] = atof( num );
        _ndata++;
        i = 0;
    }
  }
  ndata = _ndata;

  // If there is a user supplied data buffer, copy the results into it.
  if( data != NULL ){
    if( _ndata > maxdata )
      return( false );

    for( i=0; i<_ndata && i<maxdata; i++ ){
      data[i] = _data[i];
    }
  }

  return( true );
}

bool Gpib_HP_3582A::get_spectrum( double* freq, double* cha, double* chb, int& nsamp, bool& a_avail, bool& b_avail )
//------------------------------------------------------------------------------------------------------------------
// After a successful start_measurement(), read back the data and return THREE caller supplied parallel arrays.
// These arrays must all be able to hold at least 256 doubles.
//  - freq   : The frequency corresponding to each array element.
//  - cha    : The CHANNEL A data in dB or volts. Each element is the amplitude for the frequency in the
//             corresponding freq array element.
//  - chb    : The CHANNEL B data in dB or volts.
//  - nsamp  : The number of valid elements in the arrays. This will be either 128 or 256.
//  - a_avail: true if CH A data is available.
//  - b_avail: true if CH B data is available.
{
  int num=0, ago=(-1), bgo=(-1);
  int ndata;
  double data[512];

  // Zero the output arrays.
  memset( cha, 0, 256*sizeof(double) );
  memset( chb, 0, 256*sizeof(double) );

  // Read the data from the 3582A.
  if( ! read_data( ndata, data, 512 ) )
    return( false );

  // See what we should have and where it should start.
  if( _inp_chan == CHBOTH ){
    switch( _disp_chan ){
    case CHA:
      ago = 0;
      num = 128;
      break;
    case CHB:
      bgo = 0;
      num = 128;
      break;
    case CHBOTH:
      ago = 0;
      bgo = 128;
      num = 128;
      break;
    }
  }
  else{
    if( _disp_chan != _inp_chan ){
      LocalSetError( "HP3582A: Display channels and Input channels are incompatible." );
      return( false );
    }
    if( _disp_chan == CHA ){
      ago = 0;
      num = 256;
    }
    else{
      bgo = 0;
      num = 256;
    }
  }

  a_avail = ( ago >= 0 );
  b_avail = ( bgo >= 0 );
  nsamp = num;

  // Return the array contents.
  int i;
  double dfreq = ( _frhi - _frlo ) / ( num - 1 );
  for( i=0; i<num; i++ ){
    freq[i] = _frlo + i * dfreq;
  }
  if( a_avail ){
    for( i=0; i<num; i++ ){
      cha[i] = data[ago+i];
    }
  }
  if( b_avail ){
    for( i=0; i<num; i++ ){
      chb[i] = data[bgo+i];
    }
  }

  return( true );
}

bool Gpib_HP_3582A::read_annotation( int& ndata, char* text, int maxtext )
//------------------------------------------------------------------------
// Read the display annotation back from the 3582A.
{
  if( ! _hp3582a_opened ){
    LocalSetError( "HP3582A: Instrument open failed." );
    return( false );
  }

  sprintf( _gpibout, "LAN\n" );
  if( ! SendGPIB( _hp3582a ) )
    return( false );

  if( ! GetGPIB( _hp3582a ) )
    return( false );
  
  // Find how many characters we have after trimming CR, LF, trailing spaces.
  int i;
  int len = strlen( _gpibin );

  for( i=len-1; i>=0; i-- ){
    if( ( _gpibin[i] == '\n' ) || ( _gpibin[i] == '\r' ) || ( _gpibin[i] == ' ' ) )
      _gpibin[i] = 0;
    else
      break;
  }
  ndata = strlen( _gpibin );

  // Copy the returned text into the users buffer.
  if( text != NULL ){
    if( ndata >= maxtext )
      return( false );
    
    for( i=0; i<ndata; i++ )
      text[i] = _gpibin[i];
    text[i] = 0;
  }

  return( true );
}

extern "C" {
  Gpib_HP_3582A* HP_3582A_new( int a, int g )
  { return new Gpib_HP_3582A(a,g); }

  int HP_3582A_set_input_mode( Gpib_HP_3582A* s, int c )
  { return(s->set_input_mode(c)?1:0); }
  int HP_3582A_coupling( Gpib_HP_3582A* s, int c, int m )
  { return(s->coupling(c,m)?1:0); }

  int HP_3582A_set_sensitivity( Gpib_HP_3582A* s, int c, double v )
  { return(s->set_sensitivity(c,v)?1:0); }
  int HP_3582A_set_sensitivity_db( Gpib_HP_3582A* s, int c, double v )
  { return(s->set_sensitivity_db(c,v)?1:0); }

  int HP_3582A_set_frequency( Gpib_HP_3582A* s, int m, double su, double f )
  { return(s->set_frequency(m,su,f)?1:0); }

  int HP_3582A_display( Gpib_HP_3582A* s, int c, int ym )
  { return(s->display(c,ym)?1:0); }
  int HP_3582A_set_amp_ref_level_db( Gpib_HP_3582A* s, double r )
  { return(s->set_amp_ref_level_db(r)?1:0); }
  double HP_3582A_get_amp_ref_level_db(Gpib_HP_3582A* s)
  { return(s->get_amp_ref_level_db()); }

  int HP_3582A_window( Gpib_HP_3582A* s, int w )
  { return(s->window(w)?1:0); }

  int HP_3582A_start_measurement( Gpib_HP_3582A* s, int a, int c, int m )
  { return(s->start_measurement(a,c,m)?1:0); }

  int HP_3582A_read_data( Gpib_HP_3582A* s, double* d, int m )
  { int n; return(s->read_data( n, d, m )?n:(-1)); }

  int HP_3582A_read_annotation( Gpib_HP_3582A* s, char* t, int m )
  { int n; return(s->read_annotation( n, t, m )?n:(-1)); }

  int HP_3582A_get_spectrum( Gpib_HP_3582A* s, double* freq, double* cha, double* chb, 
			     int* pnsamp, int* pa_avail, int* pb_avail )
  { int nsamp;
    bool a_avail, b_avail;
    bool retstat = s->get_spectrum( freq, cha, chb, nsamp, a_avail, b_avail );
    *pnsamp = nsamp; *pa_avail = (a_avail)?1:0; *pb_avail = (b_avail)?1:0;
    return( retstat?1:0 );
  }
}

#undef BUILDING_LIBRARY

