//=====================
// HP 8903E Multimeter.
//=====================

#define BUILDING_LIBRARY
#include "GPIBlib.h"

Gpib_HP_8903E::Gpib_HP_8903E( int address, int gpibif ) : GpibInstrument( address, gpibif )
//-----------------------------------------------------------------------------------------
// Open a connection to an HP 8903E on GPIB address [address] (interface number [gpibif]).
{
  _hp8903e_opened = false;

  _err = DevOpen( gpibif, address, &_hp8903e );
  if( IsError( _err ) ){
    GpibSetError();
    return;
  }

  _global_open_count++;
  _hp8903e_opened = true;

  _function = DCV;
  _instrument_err = 0;

  // We need to set the correct terminating character for this beast.
  SetStdAttributes( _hp8903e );

  sprintf( _gpibout, "CLAU\n" );
  if( ! SendGPIB( _hp8903e ) )
    _instrument_err = 999;
}

Gpib_HP_8903E::~Gpib_HP_8903E()
//-----------------------------
// Close the connection to the HP 8903E.
{
  if( _hp8903e_opened && _global_open_count > 1 ){
    //sprintf( _gpibout, "RESET\n" );
    //SendGPIB( _hp8903e );
    DevClose( _hp8903e );
    _hp8903e_opened = false;
    _global_open_count--;
  }
}

bool Gpib_HP_8903E::check_open( void )
//------------------------------------
// Return true if open worked, else false.
// If false, set the error text appropriately.
{
  if( ! _hp8903e_opened ){
    LocalSetError( "HP8903E: Instrument open failed." );
    return( false );
  }
  return( true );
}

bool Gpib_HP_8903E::get_ac_volts( double* volts )
//-----------------------------------------------
// 
// If false, set the error text appropriately.
{
  *volts = 0.0;
  if( ! _hp8903e_opened ){
    LocalSetError( "HP8903E: Instrument open failed." );
    return( false );
  }

  // Send measure AC volts.
  sprintf( _gpibout, "M1\n" );
  if( ! SendGPIB( _hp8903e ) )
    return( false );

  // Get a string from the 3457A's output buffer.
  if( ! GetGPIB( _hp8903e ) )
    return( false );

  // Convert it to a double with no questions asked!
  *volts = atof( _gpibin );

  // See if it is an error. (Ask some questions...)
  if( *volts >= 9000000000.0 ){
    _instrument_err = (int)( ( *volts - 9000000000.0 ) / 100000.0 );
    *volts = 0.0;
    return( false );
  }

  return( true );
}

bool Gpib_HP_8903E::get_distortion( double* percent )
//---------------------------------------------------
// Return true if open worked, else false.
// If false, set the error text appropriately.
{
  *percent = 0.0;
  if( ! _hp8903e_opened ){
    LocalSetError( "HP8903E: Instrument open failed." );
    return( false );
  }

  // Send measure AC volts.
  sprintf( _gpibout, "M3L1\n" );
  if( ! SendGPIB( _hp8903e ) )
    return( false );

  // Get a string from the 3457A's output buffer.
  if( ! GetGPIB( _hp8903e ) )
    return( false );

  // Convert it to a double with no questions asked!
  *percent = atof( _gpibin );

  // See if it is an error. (Ask some questions...)
  if( *percent >= 9000000000.0 ){
    _instrument_err = (int)( ( *percent - 9000000000.0 ) / 100000.0 );
    *percent = 0.0;
    return( false );
  }

  return( true );
}


extern "C"{
  Gpib_HP_8903E* HP_8903E_new( int a, int g )
  { return new Gpib_HP_8903E(a,g); }
  int HP_8903E_check_open( Gpib_HP_8903E* s )
  { return(s->check_open()?1:0); }

  double HP_8903E_get_ac_volts( Gpib_HP_8903E* s, double* volts )
  { double m; s->get_ac_volts(&m); return(m); }
  double HP_8903E_get_distortion( Gpib_HP_8903E* s, double* percent )
  { double m; s->get_distortion(&m); return(m); }
}

#undef BUILDING_LIBRARY
