//==================================
// Sony/Tektronix RTD-710A Digitizer
//==================================

#define BUILDING_LIBRARY
#include "GPIBlib.h"

Gpib_TEK_RTD710A::Gpib_TEK_RTD710A( int address, int gpibif ) : GpibInstrument( address, gpibif )
//-----------------------------------------------------------------------------------------------
// Open a connection to a Sony/Tektronix RTD-710A on GPIB address [address] (interface number [gpibif]).
{
  _rtd710a_opened = false;

  _err = DevOpen( gpibif, address, &_rtd710a );
  if( IsError( _err ) ){
    GpibSetError();
    return;
  }

  _global_open_count++;
  _rtd710a_opened = true;

  _instrument_err = 0;

  // We need to set the correct terminating character for this beast.
  SetStdAttributes( _rtd710a );

  // Send a command to initialize the instrument.
  //sprintf( _gpibout, "INIT ALL\n" );
  //if( ! SendGPIB( _rtd710a ) )
  //  _instrument_err = 999;

  // Turn off almost all SRQs ... except WRI - acquisition complete.
  // Not that we use any of this stuff ... yet!
  sprintf( _gpibout, "RQS ON;CER OFF;EXW OFF;INR OFF;OPC OFF;USE OFF;WRI ON\n" );
  if( ! SendGPIB( _rtd710a ) )
    _instrument_err = 999;

  // Start looking for SRQ events. OR NOT!
  //viEnableEvent( _rtd710a, VI_EVENT_SERVICE_REQ, VI_QUEUE, VI_NULL );
}

Gpib_TEK_RTD710A::~Gpib_TEK_RTD710A()
//-----------------------------------
// Close the connection to the Sony/Tektronix RTD-710A.
{
  if( _rtd710a_opened && _global_open_count > 1 ){
    DevClose( _rtd710a );
    _rtd710a_opened = false;
    _global_open_count--;
  }
}

// Vertical axis

bool Gpib_TEK_RTD710A::range( int chan, double volts )
//----------------------------------------------------
// Set the full scale voltage range of channel [chan] to [volts].
{
  char* chstr;

  switch( chan ){
  case CH1:
    chstr = (char*)"CH1";
    break;
  case CH2:
    chstr = (char*)"CH2";
    break;
  default:
    return( false );
  }

  sprintf( _gpibout, "%s RAN:%.2f\n", chstr, round_volts( volts ) );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

double Gpib_TEK_RTD710A::round_volts( double volts )
//--------------------------------------------------
// Return the nearest valid voltage >= volts.
{
  double tp[] = {1.00,1.25,1.60,2.00,2.50,3.20,4.00,5.00,6.20,8.00};
  int ntp = sizeof(tp)/sizeof(double);
  int i;

  if( volts < 0.1 )
    return( 0.1 );
  if( volts < 1.0 ){
    for( i=0; i<ntp; i++ ){
      if( volts <= ( 0.1 * tp[i] ) )
        return( 0.1 * tp[i] );
    }
  }
  if( volts < 10.0 ){
    for( i=0; i<ntp; i++ ){
      if( volts <= tp[i] )
        return( tp[i] );
    }
  }
  if( volts < 100.0 ){
    for( i=0; i<ntp; i++ ){
      if( volts <= ( 10 * tp[i] ) )
        return( 10 * tp[i] );
    }
  }
  if( volts < 500.0 ){
    for( i=0; i<ntp; i++ ){
      if( volts <= ( 100 * tp[i] ) )
        return( 100 * tp[i] );
    }
  }
  return( 500.0 );
}

bool Gpib_TEK_RTD710A::offset( int chan, double volts )
//-----------------------------------------------------
// Set the offset for channel [chan] to [volts].
{
  char* chstr;

  switch( chan ){
  case CH1:
    chstr = (char*)"CH1";
    break;
  case CH2:
    chstr = (char*)"CH2";
    break;
  default:
    return( false );
  }

  sprintf( _gpibout, "%s OFF:%.2f\n", chstr, round_volts( volts ) );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

bool Gpib_TEK_RTD710A::coupling( int chan, int coupling )
//-------------------------------------------------------
// Set the coupling mode for channel [chan].
{
  char* chstr;
  char* mode;

  switch( chan ){
  case CH1:
    chstr = (char*)"CH1";
    break;
  case CH2:
    chstr = (char*)"CH2";
    break;
  default:
    return( false );
  }

  switch( coupling ){
  case AC_COUPLED:
    mode = (char*)"AC";
    break;
  case DC_COUPLED:
    mode = (char*)"DC";
    break;
  case GND_INPUT:
    mode = (char*)"GND";
    break;
  }

  sprintf( _gpibout, "%s COU:%s\n", chstr, mode );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

bool Gpib_TEK_RTD710A::bw_limit( bool yes )
//-----------------------------------------
// Turn 20 MHz bandwidth limit on or off.
{
  if( yes )
    sprintf( _gpibout, "BWL ON\n" );
  else
    sprintf( _gpibout, "BWL OFF\n" );

  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

bool Gpib_TEK_RTD710A::ch1_only( bool yes )
//-----------------------------------------
// Go to single channel mode. This allows fast sampling (200 MSamples/sec).
{
  if( yes )
    sprintf( _gpibout, "VMO CH1\n" );
  else
    sprintf( _gpibout, "VMO DUA\n" );

  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

// Horizontal axis

bool Gpib_TEK_RTD710A::sample_interval( double time )
//---------------------------------------------------
// Set the sample interval in seconds.
{
  // Convert time to the next bigger time with single digit integer mantissa.
  double exponent = log10( time );
  int iexponent = (int)floor(exponent);
  double mantissa = pow( 10.0, ( exponent - iexponent ) );
  int imantissa = (int)ceil(mantissa);

  sprintf( _gpibout, "SAM INT:%dE%d\n", imantissa, iexponent );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

bool Gpib_TEK_RTD710A::samples( int samples )
//-------------------------------------------
// Set the number of samples to record.
// Legal values are powers of 2 from 1024 to 131072 (262144 in CH1 only, high speed mode).
{
  int i;
  bool goodvalue = false;

  // Check samples is valid.
  for( i=10; i<=17; i++ ){  // 2**17 = 131072
    if( samples == ( 1 << i ) ){
      goodvalue = true;
      break;
    }
  }
  if( ! goodvalue )
    return( false );

  sprintf( _gpibout, "LEN %d\n", samples );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

bool Gpib_TEK_RTD710A::sample_hold( bool yes )
//--------------------------------------------
// Stop acquisition after the current cycle completes if [yes].
// Start acquisition cycles again if not [yes].
{
  sprintf( _gpibout, "HOL %s\n", ( yes ? "NEX" : "RES" ) );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  return( true );
}

// Triggering
bool Gpib_TEK_RTD710A::trigger_mode( int mode )
{
  return( true );
}

bool Gpib_TEK_RTD710A::trigger_source( int source )
{
  return( true );
}

bool Gpib_TEK_RTD710A::trigger_coupling( int coupling )
{
  return( true );
}

bool Gpib_TEK_RTD710A::trigger_slope( int slope )
{
  return( true );
}

bool Gpib_TEK_RTD710A::trigger_level( double percent, int which )
{
  return( true );
}

// Recording
bool Gpib_TEK_RTD710A::record_to( int location )
{
  return( true );
}

bool Gpib_TEK_RTD710A::record_mode( int mode, int times )
{
  return( true );
}

// Display

bool Gpib_TEK_RTD710A::display_from( int location, int chan )
{
  return( true );
}

bool Gpib_TEK_RTD710A::display_hzoom( int zoom )
{
  return( true );
}

bool Gpib_TEK_RTD710A::display_vzoom( int zoom )
{
  return( true );
}

bool Gpib_TEK_RTD710A::display_vpos( int pos )
{
  return( true );
}


// Get waveform

bool Gpib_TEK_RTD710A::record_wave( int waitms )
//----------------------------------------------
// Record a single waveform, at the previously selected number of samples, sample interval,
// record location, number of channels, etc.
// Wait at most [waitms] milliseconds to get a trigger and record the signal.
// NOTE: If the RTD is left storing data into a location while it is being read back by the
//       controller, there WILL be checksum errors! Sampling must be suspended prior to reading
//       back data. So we do a sort of "single shot" record here. Using SRQs and all sorts!
{
  // Try to do an acquisition and wait for it to complete.
  // Single Trigger Mode:
  sprintf( _gpibout, "TRI MOD:SGL\n" );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  // Start sampling... Hold Reset.
  sprintf( _gpibout, "HOL RES\n" );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  // Stop sampling when current acquisition cycle ends... Hold Next.
  sprintf( _gpibout, "HOL NEX\n" );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  sleep(1);

#if 1
  // Wait for an SRQ. We want code 208... Do this in-line.
  for(;;){
    char status;
    int ev = 0;

    if( ! WaitForSRQ( waitms ) ){
      LocalSetError( "TEK_RTD710A: Error waiting for SRQ event." );
      ResetGPIB( _rtd710a );
      return( false );
    }

    if( ! ReadSTB( _rtd710a, &status ) ){
      LocalSetError( "TEK_RTD710A: Error reading status byte." );
      ResetGPIB( _rtd710a );
      return( false );
    }

    else{
      printf( "status = %d (%d unsigned)\n", status, (unsigned char)status );
      if( (unsigned char)status == 208 ){
        ev = get_int_query_sep( "EVE?", ' ' );
        printf(" event=%d\n", ev );
      }
      if( ev == 750 )
        break;
    }
  } // Eternal loop!
#endif

  return( true );
}

bool Gpib_TEK_RTD710A::read_wave( int location, int chan, double* wave, int& npoints, double& tstep )
//---------------------------------------------------------------------------------------------------
// Read a waveform back from the RTD-710A. The waveform is stored in [location] and we want channel
// [chan] from it. Return the wave in [wave] and the point count in [npoints]. Also return the time
// step between samples in [tstep].
// If [wave] is NULL, just return [npoints] - use this to dynamically allocate buffer for [wave].
{
  int start, bytesread;
  double ymult;

  // Get the number of points to read. If [wave] is NULL, just return.
  // This is always going to be whatever the record length is set to (for now, anyway).
  npoints = get_int_query_sep( "LEN?", ' ' );
  if( wave == NULL )
    return( true );

  // Set the channel, location and count we want to transfer in a DATA command.
  sprintf( _gpibout, "DAT CHA:CH%d,LOC:%d,COU:%d\n", chan, location, npoints );
  if( ! SendGPIB( _rtd710a ) )
    return( false );
  
  // Get the trigger offset implied by the above (?), as we will need this for another DATa spec. command.
  start = get_int_query( "WFM? PT.O" );

  // Values to scale the 10 bit data and sample point interval to volts and time.
  ymult = get_double_query( "WFM? YMU" );
  tstep = get_double_query( "WFM? XIN" );
  ymult = ( 2.0 * ymult ) / 1024.0;
  //printf( "npoints=%d, ymult=%f, tstep=%f\n", npoints, ymult, tstep );
  
  // Send another DATa command. This sets the mysterious "start" value as well as encoding mode and transfer
  // block size (1024 bytes).
  sprintf( _gpibout, "DAT CHA:CH%d,LOC:%d,STA:%d,COU:%d,BFO:BIN,BSI:1024\n", chan, location, start, npoints );
  if( ! SendGPIB( _rtd710a ) )
    return( false );
  
  // Try sending a CURVE? command ... stand well back!
  sprintf( _gpibout, "CURVE?\n" );
  if( ! SendGPIB( _rtd710a ) )
    return( false );

  // Get the curve data - all of it! The _gpibin buffer needs to be big enough to hold up to 262144*2+10 bytes.
  ClearGPIBin();
  if( ! GetBinaryGPIB( _rtd710a, GPIB_IL, bytesread ) ){
    LocalSetError( "TEK_RTD710A: Failed to get curve data." );
    return( false );
  }
  //printf( "Read %d bytes.\n", bytesread );
  
  // See if the response is hopeful.
  if( ! strncmp( _gpibin, "CURVE", 5 ) ){
    int i, bytecount, pointcount, sum, checksum;
    unsigned char *din, *pin;
    
    // Point din to the start of the first data block.
    din = (unsigned char*)&_gpibin[7];
    
    // Loop over blocks.
    int iblock, ipt;
    int nblocks = npoints / 1024;
    for( iblock=0,ipt=0; iblock<nblocks; iblock++ ){

      // Get number of points in the block and start to form local checksum.
      sum = din[0] + din[1];
      bytecount = 256 * din[0] + din[1];
      pointcount = bytecount / 2;
      //printf( "block=%d/%d, bytecount=%d, pointcount=%d\n", iblock+1, nblocks, bytecount, pointcount );

      // Assemble each point from two bytes. Convert to voltage. Continue to form checksum.
      for( i=0,pin=&din[2]; i<pointcount; i++,ipt++ ){
        block[i] = 256 * *pin;
        sum += *pin++;
        block[i] += *pin;
        sum += *pin++;
        //printf( "%6d %5d\n", i, block[i] );
        //if( (i+1)%20 == 0 ){
        //  printf( "PAUSE> "); fflush(stdout); getchar();
        //}
        wave[ipt] = ymult * ( block[i] - 512 );
      } // Over points in this block.

      // Check our checksum is the same as the RTD-710A's checksum.
      sum = ( 256 - ( sum & 255 ) ) & 255;
      checksum = *pin++;
      //printf( "sum=%d, checksum=%d\n", sum, checksum );
      if( sum != checksum ){
        printf( "sum=%d, checksum=%d\n", sum, checksum );
        LocalSetError( "TEK_RTD710A: Bad checksum on waveform transfer." );
        ResetGPIB( _rtd710a );
        return( false );
      }

      // Move din to where the start of the next data block will be (if there is one).
      din = pin + 2; // skip ",%"
    } // Over blocks
    
  } // Got CURVE back
  
  return( true );
}

int Gpib_TEK_RTD710A::get_int_query( const char* query )
//------------------------------------------------
// Send [query]to the RTD-710A, expecting a single integer back.
// Parse out that integer and return it.
{
  return( get_int_query_sep( query, ':' ) );
}

int Gpib_TEK_RTD710A::get_int_query_sep( const char* query, char sep )
//--------------------------------------------------------------
// Send [query]to the RTD-710A, expecting a single integer back.
// Parse out that integer and return it.
{
  char* ipos;

  if( ! QueryGPIB( _rtd710a, query ) )
    return( false );
  ipos = strrchr( _gpibin, sep );
  if( ipos == NULL )
    return( 0 );
  return( atoi( ipos+1 ) );
}

double Gpib_TEK_RTD710A::get_double_query( const char* query )
//------------------------------------------------------
// Send [query]to the RTD-710A, expecting a single double back.
// Parse out that double and return it.
{
  char* ipos;

  if( ! QueryGPIB( _rtd710a, query ) )
    return( false );
  ipos = strrchr( _gpibin, ':' );
  if( ipos == NULL )
    return( 0 );
  return( atof( ipos+1 ) );
}

extern "C" {
  Gpib_TEK_RTD710A* TEK_RTD710A_new(int a, int g)
  {return new Gpib_TEK_RTD710A(a,g);}

  int TEK_RTD710A_range(Gpib_TEK_RTD710A* s, int chan, double volts)
  {return(s->range(chan, volts)?1:0);}

  int TEK_RTD710A_offset(Gpib_TEK_RTD710A* s, int chan, double volts)
  {return(s->offset(chan, volts)?1:0);}

  int TEK_RTD710A_coupling(Gpib_TEK_RTD710A* s, int chan, int coupling)
  {return(s->coupling(chan, coupling)?1:0);}

  int TEK_RTD710A_bw_limit(Gpib_TEK_RTD710A* s, int yes)
  {return(s->bw_limit(yes==1)?1:0);}

  int TEK_RTD710A_ch1_only(Gpib_TEK_RTD710A* s, int yes)
  {return(s->ch1_only(yes==1)?1:0);}

  int TEK_RTD710A_sample_interval(Gpib_TEK_RTD710A* s, double time)
  {return(s->sample_interval(time)?1:0);}

  int TEK_RTD710A_samples(Gpib_TEK_RTD710A* s, int samples)
  {return(s->samples(samples)?1:0);}

  int TEK_RTD710A_sample_hold(Gpib_TEK_RTD710A* s, int yes)
  {return(s->sample_hold(yes==1)?1:0);}

  int TEK_RTD710A_record_wave(Gpib_TEK_RTD710A* s, int waitms)
  {return(s->record_wave(waitms)?1:0);}

  int TEK_RTD710A_read_wave(Gpib_TEK_RTD710A* s, int location, int chan, double* wave, 
			     int* pnpoints, double* ptstep) 
  { int npoints;
    double tstep;
    bool retstat = s->read_wave( location, chan, wave, npoints, tstep );
    *pnpoints = npoints;
    *ptstep = tstep;
    return( retstat?1:0 );
  }
}

#undef BUILDING_LIBRARY
