#ifndef __GPIBLIB_H__
#define __GPIBLIB_H__

// GPIBlib.h - Header for a small GPIB instrument control library
// ==============================================================
//
// Nick Glazzard December 2005 - Windows VISA version.
//               December 2012 - Linux GPIBlib version.
// ----------------------------------------------------

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <math.h>
#include <stdarg.h>
#include <gpib/ib.h>

#define DEBUG_BUILD 0

#if DEBUG_BUILD
#define ENTER fprintf( stderr, "Entered: %s\n", __func__ )
#define LEAVE fprintf( stderr, "Leaving: %s\n", __func__ )
#else
#define ENTER
#define LEAVE
#endif

#define GpibHandle int
#define GpibStatus int

static int _global_open_count = 0;   // Global opened instruments count.

// Try defining standard multipliers

#define MS   *1e-3
#define US   *1e-6
#define NS   *1e-9
#define PS   *1e-12

#define MV   *1e-3
#define UV   *1e-6
#define NV   *1e-9

// Channel codes

#define CH1           1  // Channel 1
#define CH2           2  // Channel 2

#define CHA           1  // Channel A
#define CHB           2  // Channel B
#define CHBOTH        3  // Both channels.

// Coupling modes

#define AC_COUPLED    1  // AC coupling
#define DC_COUPLED    2  // DC coupling
#define GND_INPUT     3  // Ground the input

#define AC_LF_REJECT  4  // Trigger AC LF reject.
#define AC_HF_REJECT  5  // Trigger AC HF reject

// Trigger things

#define TRIG_LEVEL_1  1  // First trigger level (%)
#define TRIG_LEVEL_2  2  // Second trigger level (%)

#define SLOPE_POS     1  // Trigger slope positive.
#define SLOPE_NEG     2  // Trigger slope negative.
#define SLOPE_BI      3  // Trigger bi-slope ??


// Static buffer lengths.

#define GPIB_NL 20       // Gpib instrument + interface name length
#define GPIB_EL 256      // Error string length.
#define GPIB_OL (1<<20)  // GPIB output buffer length. 1MByte!
#define GPIB_IL (1<<20)  // GPIB input buffer length.  1MByte!
#define GPIB_MSL (1<<20) // GPIB diagnostic message buffer length.

// GpibInstrument : Stuff common to all instruments.;

class GpibInstrument
{
public:
  GpibInstrument( int address, int gpibif=0 );
  virtual ~GpibInstrument();

  int DevOpen( int brd, int dev, GpibHandle* sess );
  int DevClose( GpibHandle sess );
  char* GetLastError( void );
  int IsError( int status );
  bool SendGPIB( GpibHandle inst );
  bool GetGPIB( GpibHandle inst );
  bool GetBinaryGPIB( GpibHandle inst, int maxbytes, int& nread );
  bool QueryGPIB( GpibHandle inst, const char* query );
  bool SetStdAttributes( GpibHandle inst, char termchar=0x0a );
  bool ClearGPIBin( void );
  bool ResetGPIB( GpibHandle inst );
  bool ClearDevice( GpibHandle inst );
  bool WaitForSRQ( float msecs );
  bool ReadSTB( GpibHandle inst, char* status );
  void ResetBus( void );
  bool SetInterfaceTimeOut( float msecs );
  void Notify( const char* fmt, ... );
  virtual void notify_out( void );

  void GpibSetError( void );
  void LocalSetError( const char* errtext );
  void DisplayStatus( int status );

  void set_print_error( bool yes );
  void set_print_gpib( bool yes );
  void set_display_status( bool yes );
  void set_dump_gpib( bool yes );
  void dump_gpibin( int n );

public:
  GpibHandle _rsm;             // Session handle.
  GpibStatus  _err;            // Gpib call error code.
  bool _haveerror;             // An error has occurred.
  bool _printerror;            // Display error messages.
  bool _printgpib;             // Display all GPIB traffic.
  bool _dispstatus;            // Decode and display status bits after every call.
  bool _dumpgpibin;            // Display a representation of binary data read.
  int _address;                // Instrument address.
  int _gpibif;                 // GPIB interface card number. (0 for /dev/gpib0, etc).
  bool _termcharen;            // Termination character enabled flag.
  char _termchar;              // Termination character value.
  char _namestring[GPIB_NL];   // Instrument + Interface name. (Not really used anymore).
  char _errstring[GPIB_EL];    // Gpib call error text.
  char* _gpibout;              // Output string.
  char* _gpibin;               // Input string.
  char* _message;              // Message string.
};

// Specific instrument includes.

#include "GPIBlib_HP_3457A.h"
#include "GPIBlib_HP_3325A.h"
#include "GPIBlib_HP_8903E.h"
#include "GPIBlib_HP_3582A.h"
#include "GPIBlib_TEK_RTD710A.h"

#endif // __GPIBLIB_H__
